package uk.ac.ox.cs.goalDriven;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.chaseBench.ChaseInstance;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.parser.CSVParser;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class Utilities {

        public static Rule applyTreeDecomposition(Rule query, Set<Rule> rules, String scenarioName,
                        String scenarioNameSuffix) {
                if (query.getHeadAtom(0).getPredicate().getName().equals("q16") && scenarioName.equals("deep")
                                && scenarioNameSuffix.equals("200")) {
                        /*
                         * q16(?m262004_c0) <-
                         * m262004(?m262004_c0,?m262004_c1,?m262004_c2,?m262004_c3),
                         * m206004(?m262004_c0,?m206004_c1,?m206004_c2,?m206004_c3),
                         * m146004(?m262004_c0,?m146004_c1,?m146004_c2,?m146004_c3),
                         * m297004(?m297004_c0,?m262004_c0,?m297004_c2,?m297004_c3),
                         * m280004(?m280004_c0,?m262004_c0,?m280004_c2,?m280004_c3),
                         * m78004(?m262004_c0,?m78004_c1,?m78004_c2,?m78004_c3),
                         * m289004(?m289004_c0,?m262004_c0,?m289004_c2,?m289004_c3),
                         * m196004(?m262004_c0,?m196004_c1,?m196004_c2,?m196004_c3) .
                         */

                        Atom q16 = Atom.create(Predicate.create("q16"), new Term[] { Variable.create("X") });
                        Atom q16_1 = Atom.create(Predicate.create("q16_1"), new Term[] { Variable.create("X") });
                        Atom q16_2 = Atom.create(Predicate.create("q16_2"), new Term[] { Variable.create("X") });
                        Atom q16_3 = Atom.create(Predicate.create("q16_3"), new Term[] { Variable.create("X") });
                        Atom q16_4 = Atom.create(Predicate.create("q16_4"), new Term[] { Variable.create("X") });
                        Atom q16_5 = Atom.create(Predicate.create("q16_5"), new Term[] { Variable.create("X") });
                        Atom q16_6 = Atom.create(Predicate.create("q16_6"), new Term[] { Variable.create("X") });
                        Atom q16_7 = Atom.create(Predicate.create("q16_7"), new Term[] { Variable.create("X") });

                        Atom m196004 = Atom.create(Predicate.create("m196004"), new Term[] { Variable.create("X"),
                                        Variable.create("D1"), Variable.create("D2"), Variable.create("D3") });
                        Atom m289004 = Atom.create(Predicate.create("m289004"), new Term[] { Variable.create("D4"),
                                        Variable.create("X"), Variable.create("D5"), Variable.create("D6") });
                        Atom m78004 = Atom.create(Predicate.create("m78004"), new Term[] { Variable.create("X"),
                                        Variable.create("D7"), Variable.create("D8"), Variable.create("D9") });
                        Atom m280004 = Atom.create(Predicate.create("m280004"), new Term[] { Variable.create("D10"),
                                        Variable.create("X"), Variable.create("D11"), Variable.create("D12") });
                        Atom m297004 = Atom.create(Predicate.create("m297004"), new Term[] { Variable.create("D13"),
                                        Variable.create("X"), Variable.create("D14"), Variable.create("D15") });
                        Atom m146004 = Atom.create(Predicate.create("m146004"), new Term[] { Variable.create("X"),
                                        Variable.create("D16"), Variable.create("D17"), Variable.create("D18") });
                        Atom m206004 = Atom.create(Predicate.create("m206004"), new Term[] { Variable.create("X"),
                                        Variable.create("D19"), Variable.create("D20"), Variable.create("D21") });
                        Atom m262004 = Atom.create(Predicate.create("m262004"), new Term[] { Variable.create("X"),
                                        Variable.create("D22"), Variable.create("D23"), Variable.create("D24") });

                        query = Rule.create(q16, new Atom[] { m262004, q16_7 });
                        rules.add(Rule.create(q16_7, new Atom[] { m206004, q16_6 }));
                        rules.add(Rule.create(q16_6, new Atom[] { m146004, q16_5 }));
                        rules.add(Rule.create(q16_5, new Atom[] { m297004, q16_4 }));
                        rules.add(Rule.create(q16_4, new Atom[] { m280004, q16_3 }));
                        rules.add(Rule.create(q16_3, new Atom[] { m78004, q16_2 }));
                        rules.add(Rule.create(q16_2, new Atom[] { m289004, q16_1 }));
                        rules.add(Rule.create(q16_1, new Atom[] { m196004 }));
                } else if (query.getHeadAtom(0).getPredicate().getName().equals("q18") && scenarioName.equals("deep")
                                && scenarioNameSuffix.equals("200")) {
                        /*
                         * q18(?X,?Y,?Z,?W) <-
                         * m233004(?X,?m233004_c1,?Z,?Y),
                         * m199004(?X,?W,?m199004_c2,?m199004_c3),
                         * m111004(?X,?m111004_c1,?m111004_c2,?m111004_c3),
                         * m192004(?X,?m192004_c1,?m192004_c2,?m192004_c3),
                         * m158004(?X,?m158004_c1,?m158004_c2,?m158004_c3),
                         * m248004(?X,?m248004_c1,?m248004_c2,?m248004_c3),
                         * m6004(?m6004_c0,?X,?m6004_c2,?m6004_c3),
                         * m209004(?X,?m209004_c1,?m209004_c2,?m209004_c3) .
                         */

                        Atom q18 = Atom.create(Predicate.create("q18"),
                                        new Term[] { Variable.create("X"), Variable.create("Y"),
                                                        Variable.create("Z"), Variable.create("W") });
                        Atom q18_1 = Atom.create(Predicate.create("q18_1"), new Term[] { Variable.create("X") });
                        Atom q18_2 = Atom.create(Predicate.create("q18_2"), new Term[] { Variable.create("X") });
                        Atom q18_3 = Atom.create(Predicate.create("q18_3"), new Term[] { Variable.create("X") });
                        Atom q18_4 = Atom.create(Predicate.create("q18_4"), new Term[] { Variable.create("X") });
                        Atom q18_5 = Atom.create(Predicate.create("q18_5"), new Term[] { Variable.create("X") });
                        Atom q18_6 = Atom.create(Predicate.create("q18_6"), new Term[] { Variable.create("X") });
                        Atom q18_7 = Atom.create(Predicate.create("q18_7"),
                                        new Term[] { Variable.create("X"), Variable.create("W") });

                        Atom m233004 = Atom.create(Predicate.create("m233004"), new Term[] { Variable.create("X"),
                                        Variable.create("D1"), Variable.create("Z"), Variable.create("Y") });
                        Atom m199004 = Atom.create(Predicate.create("m199004"), new Term[] { Variable.create("X"),
                                        Variable.create("W"), Variable.create("D5"), Variable.create("D6") });
                        Atom m111004 = Atom.create(Predicate.create("m111004"), new Term[] { Variable.create("X"),
                                        Variable.create("D7"), Variable.create("D8"), Variable.create("D9") });
                        Atom m192004 = Atom.create(Predicate.create("m192004"), new Term[] { Variable.create("X"),
                                        Variable.create("D10"), Variable.create("D11"), Variable.create("D12") });
                        Atom m158004 = Atom.create(Predicate.create("m158004"), new Term[] { Variable.create("X"),
                                        Variable.create("D13"), Variable.create("D14"), Variable.create("D15") });
                        Atom m248004 = Atom.create(Predicate.create("m248004"), new Term[] { Variable.create("X"),
                                        Variable.create("D16"), Variable.create("D17"), Variable.create("D18") });
                        Atom m6004 = Atom.create(Predicate.create("m6004"), new Term[] { Variable.create("D19"),
                                        Variable.create("X"), Variable.create("D20"), Variable.create("D21") });
                        Atom m209004 = Atom.create(Predicate.create("m209004"), new Term[] { Variable.create("X"),
                                        Variable.create("D22"), Variable.create("D23"), Variable.create("D24") });

                        query = Rule.create(q18, new Atom[] { m233004, q18_7 });
                        rules.add(Rule.create(q18_7, new Atom[] { m199004, q18_6 }));
                        rules.add(Rule.create(q18_6, new Atom[] { m111004, q18_5 }));
                        rules.add(Rule.create(q18_5, new Atom[] { m192004, q18_4 }));
                        rules.add(Rule.create(q18_4, new Atom[] { m158004, q18_3 }));
                        rules.add(Rule.create(q18_3, new Atom[] { m248004, q18_2 }));
                        rules.add(Rule.create(q18_2, new Atom[] { m6004, q18_1 }));
                        rules.add(Rule.create(q18_1, new Atom[] { m209004 }));
                } else if (query.getHeadAtom(0).getPredicate().getName().equals("q20") && scenarioName.equals("deep")
                                && scenarioNameSuffix.equals("200")) {
                        /*
                         * q20(?X,?Y,?Z) <-
                         * m297004(?Y,?X,?Z,?m297004_c3),
                         * 
                         * m195004(?X,?m195004_c1,?m195004_c2,?m195004_c3),
                         * m207004(?X,?m207004_c1,?m207004_c2,?m207004_c3),
                         * 
                         * m274004(?Y,?m274004_c1,?m274004_c2,?m274004_c3),
                         * m247004(?Y,?m247004_c1,?m247004_c2,?m247004_c3),
                         * m299004(?m299004_c0,?Y,?m299004_c2,?m299004_c3),
                         * m165004(?Y,?m165004_c1,?m165004_c2,?m165004_c3),
                         * m1004(?m1004_c0,?Y,?m1004_c2,?m1004_c3) .
                         */

                        /*
                         * q20(?X,?Y,?Z) <- m297004(?Y,?X,?Z,?m297004_c3), q20_7(X), q20_5(Y)
                         * q20_7(X) <- m195004(?X,?m195004_c1,?m195004_c2,?m195004_c3), q20_6(X)
                         * q20_6(X) <- m207004(?X,?m207004_c1,?m207004_c2,?m207004_c3)
                         *
                         * q20_5(Y) <- m274004(?Y,?m274004_c1,?m274004_c2,?m274004_c3), q20_4(Y)
                         * q20_4(Y) <- m247004(?Y,?m247004_c1,?m247004_c2,?m247004_c3), q20_3(Y)
                         * q20_3(Y) <- m299004(?m299004_c0,?Y,?m299004_c2,?m299004_c3), q20_2(Y)
                         * q20_2(Y) <- m165004(?Y,?m165004_c1,?m165004_c2,?m165004_c3), q20_1(Y)
                         * q20_1(Y) <- m1004(?m1004_c0,?Y,?m1004_c2,?m1004_c3) .
                         */

                        Atom q20 = Atom.create(Predicate.create("q20"),
                                        new Term[] { Variable.create("X"), Variable.create("Y"),
                                                        Variable.create("Z") });
                        Atom q20_1 = Atom.create(Predicate.create("q20_1"), new Term[] { Variable.create("Y") });
                        Atom q20_2 = Atom.create(Predicate.create("q20_2"), new Term[] { Variable.create("Y") });
                        Atom q20_3 = Atom.create(Predicate.create("q20_3"), new Term[] { Variable.create("Y") });
                        Atom q20_4 = Atom.create(Predicate.create("q20_4"), new Term[] { Variable.create("Y") });
                        Atom q20_5 = Atom.create(Predicate.create("q20_5"), new Term[] { Variable.create("Y") });
                        Atom q20_6 = Atom.create(Predicate.create("q20_6"), new Term[] { Variable.create("X") });
                        Atom q20_7 = Atom.create(Predicate.create("q20_7"), new Term[] { Variable.create("X") });

                        Atom m297004 = Atom.create(Predicate.create("m297004"), new Term[] { Variable.create("Y"),
                                        Variable.create("X"), Variable.create("Z"), Variable.create("D1") });
                        Atom m195004 = Atom.create(Predicate.create("m195004"), new Term[] { Variable.create("X"),
                                        Variable.create("D4"), Variable.create("D5"), Variable.create("D6") });
                        Atom m207004 = Atom.create(Predicate.create("m207004"), new Term[] { Variable.create("X"),
                                        Variable.create("D7"), Variable.create("D8"), Variable.create("D9") });

                        Atom m274004 = Atom.create(Predicate.create("m274004"), new Term[] { Variable.create("Y"),
                                        Variable.create("D10"), Variable.create("D11"), Variable.create("D12") });
                        Atom m247004 = Atom.create(Predicate.create("m247004"), new Term[] { Variable.create("Y"),
                                        Variable.create("D13"), Variable.create("D14"), Variable.create("D15") });
                        Atom m299004 = Atom.create(Predicate.create("m299004"), new Term[] { Variable.create("D16"),
                                        Variable.create("Y"), Variable.create("D17"), Variable.create("D18") });
                        Atom m165004 = Atom.create(Predicate.create("m165004"), new Term[] { Variable.create("Y"),
                                        Variable.create("D19"), Variable.create("D20"), Variable.create("D21") });
                        Atom m1004 = Atom.create(Predicate.create("m1004"), new Term[] { Variable.create("D22"),
                                        Variable.create("Y"), Variable.create("D23"), Variable.create("D24") });

                        query = Rule.create(q20, new Atom[] { m297004, q20_7, q20_5 });
                        rules.add(Rule.create(q20_7, new Atom[] { m195004, q20_6 }));
                        rules.add(Rule.create(q20_6, new Atom[] { m207004 }));

                        rules.add(Rule.create(q20_5, new Atom[] { m274004, q20_4 }));
                        rules.add(Rule.create(q20_4, new Atom[] { m247004, q20_3 }));
                        rules.add(Rule.create(q20_3, new Atom[] { m299004, q20_2 }));
                        rules.add(Rule.create(q20_2, new Atom[] { m165004, q20_1 }));
                        rules.add(Rule.create(q20_1, new Atom[] { m1004 }));
                }

                return query;
        }

        public static Set<Rule> loadQueries(ChaseInstance chaseInstance, String queryFileName) throws IOException {
                Set<Rule> queries = new LinkedHashSet<Rule>();
                if (queryFileName.length() != 0)
                        CommonParser.parse(new File(chaseInstance.m_queriesRoot, queryFileName),
                                        new InputCollector(queries, null));
                else {
                        final String TXT = ".txt";
                        final int TXTlength = TXT.length();
                        File[] queryFiles = chaseInstance.m_queriesRoot.listFiles(new FilenameFilter() {
                                public boolean accept(File dir, String name) {
                                        int length = name.length();
                                        if (length < TXTlength)
                                                return false;
                                        for (int index = 0; index < TXTlength; ++index)
                                                if (Character.toLowerCase(
                                                                name.charAt(length - TXTlength + index)) != TXT
                                                                                .charAt(index))
                                                        return false;
                                        return true;
                                }
                        });
                        for (File queryFile : queryFiles)
                                CommonParser.parse(queryFile, new InputCollector(queries, null));
                }
                return queries;
        }

        public static void loadCSVResources(ChaseInstance chaseInstance, InputProcessor inputProcessor,
                        Map<Predicate, PredicateInfo> predicateInfo) throws IOException {
                loadCSVResources(chaseInstance.m_sourceInstanceRoot, inputProcessor, predicateInfo);
        }

        public static void loadCSVResources(File sourceInstanceRoot, InputProcessor inputProcessor,
                        Map<Predicate, PredicateInfo> predicateInfo) throws IOException {
                inputProcessor.startProcessing();
                for (Predicate predicate : predicateInfo.keySet()) {
                        if (!predicateInfo.get(predicate).isIDB()) {
                                String fileName = sourceInstanceRoot + "/" + predicate.getName()
                                                + ".csv";
                                File file = new File(fileName);
                                if (file.exists()) {
                                        try (Reader reader = new FileReader(fileName)) {
                                                CSVParser.parse(reader, predicate, inputProcessor);
                                        }
                                }
                        }
                }
                inputProcessor.endProcessing();
        }

        public static Set<Rule> createAuxiliarySOAxiomsForBodyFunctions(Map<Predicate, Integer> bodyPredicates) {
                Set<Rule> rules = new LinkedHashSet<>();
                for (Entry<Predicate, Integer> entry : bodyPredicates.entrySet()) {
                        Predicate predicate = entry.getKey();
                        int arity = entry.getValue();
                        Term[] terms1 = new Term[arity];
                        Term[] terms2 = new Term[arity];
                        for (int i = 0; i < arity - 1; ++i) {
                                terms1[i] = Variable.create("X" + i);
                                terms2[i] = Variable.create("X" + i);
                        }
                        terms1[arity - 1] = Variable.create("Y");
                        terms2[arity - 1] = Variable.create("Yp");
                        rules.add(Rule.create(
                                        Atom.create(Predicate.EQUALS, Variable.create("Y"), Variable.create("Yp")),
                                        Atom.create(predicate, terms1), Atom.create(predicate, terms2)));
                }
                return rules;
        }

        public static CommonFormatProcessor printer(ChaseInstance chaseInstance, String queryName,
                        String programName)
                        throws FileNotFoundException {
                return new CommonFormatProcessor(new PrintWriter(
                                new File(chaseInstance.m_transformedProgramsRoot,
                                                queryName + "_" + programName + ".txt")));
        }
}
