package uk.ac.ox.cs.goalDriven.chaseBench;

import java.io.File;
import java.io.FileReader;
import java.io.Reader;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.parser.CSVParser;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class DebugChase {

    protected static class Result {
        int m_numberOfRules;
        int m_numberOfImportedFiles;
        long m_importDuration;
        long m_reasoningDuration;
    }
    
    protected static Result processRules(File dataRootFile, File rulesFile) throws Exception {
        Result result = new Result();
        Set<Rule> rules = new HashSet<Rule>();
        try (Reader reader = new FileReader(rulesFile)) {
            Set<Atom> facts = new HashSet<Atom>();
            InputCollector ruleCollector = new InputCollector(rules, facts);
            new CommonParser(reader).parse(ruleCollector);
        }
        result.m_numberOfRules = rules.size();
        Map<Predicate, PredicateInfo> predicateMap = CommonUtilities.getPredicateInfo(rules);
        try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(false)) {
            InputProcessor inputProcessor = jrdfoxBridge.newInputProcessor();
            long importStart = System.currentTimeMillis();
            inputProcessor.startProcessing();
            for (Predicate predicate : predicateMap.keySet()) {
                if (!predicateMap.get(predicate).isIDB()) {
                    File file = new File(dataRootFile, predicate.getName() + ".csv");
                    if (file.exists()) {
                        ++result.m_numberOfImportedFiles;
                        try (Reader reader = new FileReader(file)) {
                            CSVParser.parse(reader, predicate, inputProcessor);
                        }
                    }
                }
            }
            inputProcessor.endProcessing();
            result.m_importDuration = System.currentTimeMillis() - importStart;
            inputProcessor.startProcessing();
            for (Rule rule : rules)
                inputProcessor.processRule(rule);
            inputProcessor.endProcessing();
            long reasoningStart = System.currentTimeMillis();
            jrdfoxBridge.computeMaterialization();
            result.m_reasoningDuration = System.currentTimeMillis() - reasoningStart;
        }
        return result;
    }
    
    public static void main(String[] args) throws Exception {
        String dataRoot = "";
        String rulesRoot = "";
        
        dataRoot = "/Users/bmotik/RDFdata/chaseBench/scenarios/LUBM/data/100";
        rulesRoot = "/Users/bmotik/Downloads/comparison/";

        File dataRootFile = new File(dataRoot);
        System.out.println("Name\tRules\t(%)\tFiles\t(%)\tImport time\t(%)\tReasoning time\t(%)");
        System.out.print("full.txt");
        System.out.flush();
        Result baseline = processRules(dataRootFile, new File(rulesRoot + "full.txt"));
        System.out.print('\t');
        System.out.print(baseline.m_numberOfRules);
        System.out.print("\t100\t");
        System.out.print(baseline.m_numberOfImportedFiles);
        System.out.print("\t100\t");
        System.out.print(baseline.m_importDuration);
        System.out.print("\t100\t");
        System.out.print(baseline.m_reasoningDuration);
        System.out.println("\t100");
        for (File ruleFile : new File(rulesRoot).listFiles()) {
            if (ruleFile.getName().endsWith(".txt") && !ruleFile.getName().equals("full.txt")) {
                System.out.print(ruleFile.getName());
                System.out.flush();
                Result result = processRules(dataRootFile, ruleFile);
                System.out.print('\t');
                System.out.print(result.m_numberOfRules);
                System.out.print('\t');
                System.out.print(result.m_numberOfRules * 100 / baseline.m_numberOfRules);
                System.out.print("\t");                
                System.out.print(result.m_numberOfImportedFiles);
                System.out.print('\t');
                System.out.print(result.m_numberOfImportedFiles * 100 / baseline.m_numberOfImportedFiles);
                System.out.print("\t");                
                System.out.print(result.m_importDuration);
                System.out.print('\t');
                System.out.print(result.m_importDuration * 100 / baseline.m_importDuration);
                System.out.print("\t");                
                System.out.print(result.m_reasoningDuration);
                System.out.print('\t');
                System.out.print(result.m_reasoningDuration * 100 / baseline.m_reasoningDuration);
                System.out.println();                
            }
        }
    }

}
