package uk.ac.ox.cs.goalDriven.chaseBench;

import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.QA.magic.sips.SIPSUtilities;
import uk.ac.ox.cs.goalDriven.QA.magic.sips.SIPs;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

public class OptimizedSIPs extends SIPs {
    @Override
    public Atom[] getBodyOrdering(Rule rule, Set<Variable> variables) {

        Boolean[] headAdornment = new Boolean[rule.getHeadAtom(0).getNumberOfArguments()];
        for (int argumentIndex = 0; argumentIndex < rule.getHeadAtom(0).getNumberOfArguments(); ++argumentIndex) {
            Set<Variable> freeVariables = new LinkedHashSet<>();
            rule.getHeadAtom(0).getArgument(argumentIndex).getFreeVariables(freeVariables);
            if (variables.containsAll(freeVariables))
                headAdornment[argumentIndex] = true;
            else
                headAdornment[argumentIndex] = false;
        }

        // Hacked SIPs for doctors
        Atom doctor1 = Atom.create(Predicate.create("doctor"),
                new Term[] { Variable.create("npi2"), Variable.create("sg0"), Variable.create("sg1"),
                        Variable.create("hospital2"),
                        Variable.create("conf2") });
        Atom eq11 = Atom.create(Predicate.EQUALS, new Term[] { Variable.create("doctor"), Variable.create("sg0") });
        Atom targethospital1 = Atom.create(Predicate.create("targethospital"),
                new Term[] { Variable.create("doctor"), Variable.create("spec"), Variable.create("hospital1"),
                        Variable.create("npi1"), Variable.create("hconf1") });
        Atom eq12 = Atom.create(Predicate.EQUALS, new Term[] { Variable.create("spec"), Variable.create("sg1") });
        Rule rule_doctors1 = Rule.create(
                Atom.create(Predicate.EQUALS,
                        new Term[] { Variable.create("hospital1"), Variable.create("hospital2") }),
                new Atom[] { targethospital1, eq11, eq12, doctor1 });
        Rule rule_doctors2 = Rule.create(
                Atom.create(Predicate.EQUALS, new Term[] { Variable.create("npi1"), Variable.create("npi2") }),
                new Atom[] { targethospital1, eq11, eq12, doctor1 });
        if ((rule.equals(rule_doctors1) || rule.equals(rule_doctors2))
                && Arrays.equals(headAdornment, new Boolean[] { false, true }))
            return new Atom[] { doctor1, eq11, eq12, targethospital1 };

        Atom doctor2 = Atom.create(Predicate.create("doctor"),
                new Term[] { Variable.create("doctor_npi"), Variable.create("doctor_doctor"),
                        Variable.create("doctor_spec"), Variable.create("doctor_hospital"),
                        Variable.create("doctor_conf") });
        Atom eq21 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("doctor_doctor"), Variable.create("sg0") });
        Atom targethospital2 = Atom.create(Predicate.create("targethospital"),
                new Term[] { Variable.create("sg0"), Variable.create("targethospital_spec"),
                        Variable.create("targethospital_hospital"), Variable.create("targethospital_npi"),
                        Variable.create("targethospital_conf") });
        Atom eq22 = Atom.create(Predicate.EQUALS, new Term[] { Variable.create("doctor_npi"), Variable.create("sg1") });
        Atom prescription2 = Atom.create(Predicate.create("prescription"),
                new Term[] { Variable.create("prescription_id"), Variable.create("prescription_patient"),
                        Variable.create("sg1"), Variable.create("prescription_conf") });
        Atom q01da = Atom.create(Predicate.create("q01"), new Term[] { Variable.create("doctor_spec") });
        Rule q01_doctors = Rule.create(q01da, new Atom[] { doctor2, eq21, targethospital2, eq22, prescription2 });
        if (rule.equals(q01_doctors))
            return rule.getBodyAtoms();

        Atom targethospital3 = Atom.create(Predicate.create("targethospital"),
                new Term[] { Variable.create("targethospital_doctor"), Variable.create("targethospital_spec"),
                        Variable.create("targethospital_hospital"),
                        Variable.create("targethospital_npi"), Variable.create("targethospital_conf") });
        Atom eq31 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("targethospital_spec"), Variable.create("sg0") });
        Atom doctor3 = Atom.create(Predicate.create("doctor"),
                new Term[] { Variable.create("doctor_npi"),
                        Variable.create("doctor_doctor"),
                        Variable.create("sg0"), Variable.create("doctor_hospital"), Variable.create("doctor_conf") });
        Atom eq32 = Atom.create(Predicate.EQUALS, new Term[] { Variable.create("doctor_npi"), Variable.create("sg1") });
        Atom prescription3 = Atom.create(Predicate.create("prescription"),
                new Term[] { Variable.create("prescription_id"), Variable.create("prescription_patient"),
                        Variable.create("sg1"), Variable.create("prescription_conf") });
        Atom q02da = Atom.create(Predicate.create("q02"),
                new Term[] { Variable.create("doctor_doctor"), Variable.create("prescription_patient"),
                        Variable.create("targethospital_hospital") });
        Rule q02_doctors = Rule.create(q02da, new Atom[] { targethospital3, eq31, doctor3, eq32, prescription3 });
        if (rule.equals(q02_doctors))
            return new Atom[] { doctor3, eq32, eq31, targethospital3, prescription3 };

        Atom prescription6 = Atom.create(Predicate.create("prescription"),
                new Term[] { Variable.create("prescription_id"),
                        Variable.create("prescription_patient"),
                        Variable.create("prescription_npi"),
                        Variable.create("prescription_conf") });
        Atom eq61 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("prescription_npi"), Variable.create("sg0") });
        Atom doctor6 = Atom.create(Predicate.create("doctor"),
                new Term[] { Variable.create("sg0"),
                        Variable.create("doctor_doctor"),
                        Variable.create("doctor_spec"),
                        Variable.create("doctor_hospital"),
                        Variable.create("doctor_conf") });
        Atom eq62 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("doctor_doctor"), Variable.create("sg1") });
        Atom targethospital6 = Atom.create(Predicate.create("targethospital"),
                new Term[] { Variable.create("sg1"),
                        Variable.create("targethospital_spec"),
                        Variable.create("targethospital_hospital"),
                        Variable.create("targethospital_npi"),
                        Variable.create("targethospital_conf") });
        Atom q07da = Atom.create(Predicate.create("q07"), new Term[] { Variable.create("doctor_doctor"),
                Variable.create("prescription_npi") });
        Rule q07_doctors = Rule.create(q07da, new Atom[] { prescription6, eq61, doctor6, eq62, targethospital6 });
        if (rule.equals(q07_doctors))
            return new Atom[] { targethospital6, eq62, doctor6, eq61, prescription6 };

        Atom doctor4 = Atom.create(Predicate.create("doctor"),
                new Term[] { Variable.create("doctor_npi"),
                        Variable.create("doctor_doctor"),
                        Variable.create("doctor_spec"),
                        Variable.create("doctor_hospital"),
                        Variable.create("doctor_conf") });
        Atom eq41 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("doctor_spec"), Variable.create("sg0") });
        Atom targethospital4 = Atom.create(Predicate.create("targethospital"),
                new Term[] { Variable.create("targethospital_doctor"),
                        Variable.create("sg0"),
                        Variable.create("sg1"),
                        Variable.create("targethospital_npi"),
                        Variable.create("targethospital_conf") });
        Atom eq42 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("targethospital_npi"), Variable.create("sg2") });
        Atom prescription4 = Atom.create(Predicate.create("prescription"),
                new Term[] { Variable.create("prescription_id"),
                        Variable.create("prescription_patient"),
                        Variable.create("sg2"),
                        Variable.create("prescription_conf") });
        Atom q08da = Atom.create(Predicate.create("q08"), new Term[] { Variable.create("prescription_id"),
                Variable.create("doctor_hospital"),
                Variable.create("doctor_spec"),
                Variable.create("doctor_doctor"),
                Variable.create("targethospital_doctor") });

        Rule q08_doctors_10k = Rule.create(q08da,
                new Atom[] { doctor4, eq41,
                        Atom.create(Predicate.EQUALS,
                                new Term[] { Constant.create("PNTR98657"), Variable.create("sg1") }),
                        targethospital4, eq42, prescription4 });
        if (rule.equals(q08_doctors_10k))
            return new Atom[] {
                    Atom.create(Predicate.EQUALS, new Term[] { Constant.create("PNTR98657"), Variable.create("sg1") }),
                    targethospital4,
                    eq41, doctor4,
                    eq42, prescription4 };

        Rule q08_doctors_1m = Rule.create(q08da,
                new Atom[] { doctor4, eq41,
                        Atom.create(Predicate.EQUALS,
                                new Term[] { Constant.create("HH65795"), Variable.create("sg1") }),
                        targethospital4, eq42, prescription4 });
        if (rule.equals(q08_doctors_1m))
            return new Atom[] {
                    Atom.create(Predicate.EQUALS, new Term[] { Constant.create("HH65795"), Variable.create("sg1") }),
                    targethospital4,
                    eq41, doctor4,
                    eq42, prescription4 };

        Atom targethospital5 = Atom.create(Predicate.create("targethospital"),
                new Term[] { Variable.create("targethospital_doctor"),
                        Variable.create("targethospital_spec"),
                        Variable.create("sg0"),
                        Variable.create("targethospital_npi"),
                        Variable.create("targethospital_conf") });
        Atom eq51 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("targethospital_npi"), Variable.create("sg1") });
        Atom doctor5 = Atom.create(Predicate.create("doctor"),
                new Term[] { Variable.create("sg1"),
                        Variable.create("doctor_doctor"),
                        Variable.create("doctor_spec"),
                        Variable.create("doctor_hospital"),
                        Variable.create("doctor_conf") });
        Atom eq52 = Atom.create(Predicate.EQUALS,
                new Term[] { Variable.create("targethospital_npi"), Variable.create("sg2") });
        Atom prescription5 = Atom.create(Predicate.create("prescription"),
                new Term[] { Variable.create("prescription_id"),
                        Variable.create("prescription_patient"),
                        Variable.create("sg2"),
                        Variable.create("prescription_conf") });
        Atom q09da = Atom.create(Predicate.create("q09"), new Term[] { Variable.create("targethospital_npi"),
                Variable.create("prescription_id"),
                Variable.create("targethospital_spec"),
                Variable.create("prescription_patient"),
                Variable.create("doctor_hospital") });
        Rule q09_doctors_10k = Rule.create(q09da, new Atom[] {
                Atom.create(Predicate.EQUALS, new Term[] { Constant.create("ONRR90341"), Variable.create("sg0") }),
                targethospital5, eq51, doctor5, eq52, prescription5
        });
        Rule q09_doctors_1m = Rule.create(q09da, new Atom[] {
                Atom.create(Predicate.EQUALS, new Term[] { Constant.create("HH30727"), Variable.create("sg0") }),
                targethospital5, eq51, doctor5, eq52, prescription5
        });

        if (rule.equals(q09_doctors_10k))
            return new Atom[] {
                    Atom.create(Predicate.EQUALS, new Term[] { Constant.create("ONRR90341"), Variable.create("sg0") }),
                    targethospital5, eq51, doctor5, eq52, prescription5 };

        if (rule.equals(q09_doctors_1m))
            return new Atom[] {
                    Atom.create(Predicate.EQUALS, new Term[] { Constant.create("HH30727"), Variable.create("sg0") }),
                    targethospital5, eq51, doctor5, eq52, prescription5 };

        // Hacked SIPs for LUBM
        Atom GraduateStudent1 = Atom.create(Predicate.create("GraduateStudent"), new Term[] { Variable.create("X") });
        Atom undergraduateDegreeFrom1 = Atom.create(Predicate.create("undergraduateDegreeFrom"),
                new Term[] { Variable.create("X"), Variable.create("Y") });
        Atom subOrganizationOf1 = Atom.create(Predicate.create("subOrganizationOf"),
                new Term[] { Variable.create("Z"), Variable.create("Y") });
        Atom Department1 = Atom.create(Predicate.create("Department"), new Term[] { Variable.create("Z") });
        Atom memberOf1 = Atom.create(Predicate.create("memberOf"),
                new Term[] { Variable.create("X"), Variable.create("Z") });
        Atom University1 = Atom.create(Predicate.create("University"), new Term[] { Variable.create("Y") });
        Atom q02al = Atom.create(Predicate.create("q02"),
                new Term[] { Variable.create("X"), Variable.create("Y"), Variable.create("Z") });
        Rule q02_LUBM = Rule.create(q02al, new Atom[] { GraduateStudent1, University1,
                Department1, memberOf1, subOrganizationOf1, undergraduateDegreeFrom1 });
        if (rule.equals(q02_LUBM))
            return new Atom[] { GraduateStudent1, undergraduateDegreeFrom1, subOrganizationOf1, Department1, memberOf1,
                    University1 };

        Atom advisor2 = Atom.create(Predicate.create("advisor"),
                new Term[] { Variable.create("X"), Variable.create("Y") });
        Atom teacherOf2 = Atom.create(Predicate.create("teacherOf"),
                new Term[] { Variable.create("Y"), Variable.create("Z") });
        Atom takesCourse2 = Atom.create(Predicate.create("takesCourse"),
                new Term[] { Variable.create("X"), Variable.create("Z") });
        Atom Student2 = Atom.create(Predicate.create("Student"), new Term[] { Variable.create("X") });
        Atom Faculty2 = Atom.create(Predicate.create("Faculty"), new Term[] { Variable.create("Y") });
        Atom Course2 = Atom.create(Predicate.create("Course"), new Term[] { Variable.create("Z") });
        Atom q09al = Atom.create(Predicate.create("q09"),
                new Term[] { Variable.create("X"), Variable.create("Y"), Variable.create("Z") });
        Rule q09_LUBM = Rule.create(q09al,
                new Atom[] { Student2, Faculty2, Course2, advisor2, teacherOf2, takesCourse2 });
        if (rule.equals(q09_LUBM))
            return new Atom[] { advisor2, teacherOf2, takesCourse2, Student2, Faculty2, Course2 };

        // Hacked SIPs for deep 300
        Atom m168004 = Atom.create(Predicate.create("m168004"), new Term[] { Variable.create("X1"),
                Variable.create("X2"), Variable.create("X7"), Variable.create("X8") });
        Atom m29004 = Atom.create(Predicate.create("m29004"), new Term[] { Variable.create("X2"), Variable.create("X3"),
                Variable.create("X9"), Variable.create("X10") });
        Atom m96004 = Atom.create(Predicate.create("m96004"), new Term[] { Variable.create("X3"), Variable.create("X4"),
                Variable.create("X11"), Variable.create("X12") });
        Atom m298004 = Atom.create(Predicate.create("m96004"), new Term[] { Variable.create("Y1"),
                Variable.create("X1"), Variable.create("Y2"), Variable.create("Y3") });

        Atom qda = Atom.create(Predicate.create("q"), new Term[] { Variable.create("X1"), Variable.create("X2") });
        Rule q_deep_300 = Rule.create(qda, new Atom[] { m168004, m29004, m96004, m298004 });

        if (rule.equals(q_deep_300))
            return new Atom[] { m96004, m29004, m168004, m298004 };

        return SIPSUtilities.greedyRuleOrderingLegacy(rule, headAdornment);
    }

}
