package uk.ac.ox.cs.goalDriven.chaseBench;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.Utilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridgeUtilities;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;
import uk.ac.ox.cs.goalDriven.processors.Skolemization;

public class RunFOChase {

    public static void main(String[] args) throws Exception {
        ChaseInstance chaseInstance = new ChaseInstance();
        String scenarioNamePrefix = "";
        String scenarioName = "";
        String scenarioNameSuffix = "";
        String dataVariantName = "";
        String queryFileName = "";

        // scenarioName = "LUBM";
        // dataVariantName = "100";
        // dataVariantName = "01k";
        scenarioName = "deep";
        scenarioNameSuffix = "200";
        // scenarioNameSuffix = "300";
        // scenarioName = "Ontology-256";

        chaseInstance.initialize(scenarioNamePrefix, scenarioName, scenarioNameSuffix, dataVariantName, false);
        Set<Rule> queries = Utilities.loadQueries(chaseInstance, queryFileName);

        Set<Rule> rules = new LinkedHashSet<Rule>();
        Set<Atom> facts = new LinkedHashSet<Atom>();
        // Load the rules
        if (chaseInstance.m_stTGDsFile != null)
            CommonParser.parse(chaseInstance.m_stTGDsFile, new InputCollector(rules, facts));
        if (chaseInstance.m_tEGDsFile != null)
            CommonParser.parse(chaseInstance.m_tEGDsFile, new InputCollector(rules, facts));
        if (chaseInstance.m_tTGDsFile != null)
            CommonParser.parse(chaseInstance.m_tTGDsFile, new InputCollector(rules, facts));

        Set<Rule> skolemizedRules = new LinkedHashSet<>();
        InputProcessor chain = new Skolemization(true, true, true,
                new InputCollector(skolemizedRules, new LinkedHashSet<Atom>())) {
            protected int m_nextExistentialCounter = 0;

            protected int getNextExistentialCounter() {
                return m_nextExistentialCounter++;
            }
        };
        chain.startProcessing();
        for (Rule inputRule : rules)
            chain.processRule(inputRule);
        chain.endProcessing();

        chaseInstance.m_outputRoot.mkdirs();
        try (PrintWriter fullOutput = new PrintWriter(new File(chaseInstance.m_outputRoot, "full.txt"))) {
            CommonFormatProcessor commonFormat = new CommonFormatProcessor(fullOutput);
            commonFormat.startProcessing();
            for (Rule skolemizedRule : skolemizedRules)
                commonFormat.processRule(skolemizedRule);
            commonFormat.endProcessing();
        }

        // Apply the chase using RDFox
        try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(chaseInstance.usesEquality())) {
            Map<Predicate, PredicateInfo> predicateInfo = CommonUtilities.getPredicateInfo(skolemizedRules);
            InputProcessor inputProcessor = jrdfoxBridge.newInputProcessor();

            System.out.print("Loading the source instance... ");
            System.out.flush();
            long start = System.currentTimeMillis();
            Utilities.loadCSVResources(chaseInstance, inputProcessor, predicateInfo);
            long duration = System.currentTimeMillis() - start;
            System.out.println(duration + " ms");

            inputProcessor.startProcessing();
            for (Rule rule : skolemizedRules)
                inputProcessor.processRule(rule);
            inputProcessor.endProcessing();

            System.out.print("Applying " + +skolemizedRules.size() + " rules... ");
            System.out.flush();
            start = System.currentTimeMillis();
            jrdfoxBridge.computeMaterialization();
            duration = System.currentTimeMillis() - start;
            System.out.println(duration + " ms");

            for (Rule query : queries) {
                Atom queryAtom = query.getHeadAtom(0);
                Set<Variable> answerVariables = new LinkedHashSet<>();
                queryAtom.getFreeVariables(answerVariables);
                System.out.print("Answering query " + queryAtom.getPredicate() + " ...");
                System.out.flush();
                start = System.currentTimeMillis();
                JRDFoxBridgeUtilities.evaluateQuery(jrdfoxBridge, new ArrayList<>(answerVariables),
                        Arrays.asList(query.getBodyAtoms()),
                        (x) -> {
                            return jrdfoxBridge.parseResource(x);
                        }, true, false);
                duration = System.currentTimeMillis() - start;
                System.out.println(duration + " ms");
            }

            String modelPath = scenarioName + "/" + "model" + "/" + dataVariantName;
            new File(modelPath).mkdirs();
            Integer[] entry = CommonUtilities.saveInstanceOnDisk(jrdfoxBridge,
                    modelPath, skolemizedRules, null);
            System.out.println("Total facts " + entry[0] + "; total magic facts " + entry[1]
                    + "; total D + F facts " + entry[2] + "; useful facts " + (entry[0] - entry[1] - entry[2]));

        }

    }

}
