package uk.ac.ox.cs.goalDriven.chaseBench;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.Arrays;
import uk.ac.ox.cs.goalDriven.Utilities;
import uk.ac.ox.cs.goalDriven.QA.magic.MagicProcessor;
//import uk.ac.ox.cs.goalDriven.magicold.MagicProcessor;
//import uk.ac.ox.cs.goalDriven.magicold.sips.MaximumVariableCoverSIPs;
import uk.ac.ox.cs.goalDriven.QA.processors.Defun;
import uk.ac.ox.cs.goalDriven.QA.processors.Desingularization;
import uk.ac.ox.cs.goalDriven.QA.processors.RuleRelevance;
import uk.ac.ox.cs.goalDriven.QA.processors.RuleSubsumption;
import uk.ac.ox.cs.goalDriven.QA.processors.RuleVerifier;
import uk.ac.ox.cs.goalDriven.QA.processors.Singularization;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;
import uk.ac.ox.cs.goalDriven.processors.Skolemization;

public class RunFOPipeline {

    public static void main(String[] args) throws Exception {
        ChaseInstance chaseInstance = new ChaseInstance();
        String scenarioNamePrefix = "";
        String scenarioName = "";
        String scenarioNameSuffix = "";
        String dataVariantName = "";
        String queryFileName = "";
        TRANSFORMATION transformation = TRANSFORMATION.MAGIC;
        Boolean doFastRelevanceAnalysis = true;

        // scenarioName = "LUBM";
        // dataVariantName = "100";
        // dataVariantName = "01k";
        scenarioName = "deep";
        // scenarioNameSuffix = "200";
        scenarioNameSuffix = "300";
        // scenarioName = "Ontology-256";

        chaseInstance.initialize(scenarioNamePrefix, scenarioName, scenarioNameSuffix, dataVariantName, true);
        Set<Rule> queries = Utilities.loadQueries(chaseInstance, queryFileName);

        for (Rule query : queries) {
            Set<Rule> rules = new LinkedHashSet<Rule>();
            Set<Atom> facts = new LinkedHashSet<Atom>();
            // Load the rules
            if (chaseInstance.m_stTGDsFile != null)
                CommonParser.parse(chaseInstance.m_stTGDsFile, new InputCollector(rules, facts));
            if (chaseInstance.m_tEGDsFile != null)
                CommonParser.parse(chaseInstance.m_tEGDsFile, new InputCollector(rules, facts));
            if (chaseInstance.m_tTGDsFile != null)
                CommonParser.parse(chaseInstance.m_tTGDsFile, new InputCollector(rules, facts));

            // Apply tree decomposition
            query = Utilities.applyTreeDecomposition(query, rules, scenarioName, scenarioNameSuffix);
            rules.add(query);
            System.out.println(query.getHeadAtom(0).getPredicate());

            Set<Rule> transformedRules = transformRules(chaseInstance, query, rules, transformation,
                    doFastRelevanceAnalysis);

            // Apply the chase using RDFox
            try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(chaseInstance.usesEquality())) {
                Map<Predicate, PredicateInfo> predicateInfo = CommonUtilities.getPredicateInfo(transformedRules);
                InputProcessor inputProcessor = jrdfoxBridge.newInputProcessor();

                System.out.print("Loading the source instance... ");
                System.out.flush();
                long start = System.currentTimeMillis();
                Utilities.loadCSVResources(chaseInstance, inputProcessor, predicateInfo);
                long duration = System.currentTimeMillis() - start;
                System.out.println(duration + " ms");

                inputProcessor.startProcessing();
                for (Rule rule : transformedRules)
                    inputProcessor.processRule(rule);
                inputProcessor.endProcessing();

                System.out.print("Applying " + +transformedRules.size() + " rules... ");
                System.out.flush();
                start = System.currentTimeMillis();
                jrdfoxBridge.computeMaterialization();
                duration = System.currentTimeMillis() - start;
                System.out.println(duration + " ms");

                String modelPath = scenarioName + "/" + transformation.name() + "/"
                        + query.getHeadAtom(0).getPredicate().getName() + "/"
                        + dataVariantName;
                new File(modelPath).mkdirs();
                Integer[] entry = CommonUtilities.saveInstanceOnDisk(jrdfoxBridge,
                        modelPath, transformedRules, Arrays.asList(query.getHeadAtom(0).getPredicate()));
                System.out.println("Total facts " + entry[0] + "; total magic facts " + entry[1]
                        + "; total D + F facts " + entry[2] + "; useful facts " + (entry[0] - entry[1] - entry[2]));
                System.out.println();
                System.out.println();

            }
        }
    }

    public static enum TRANSFORMATION {
        MAGIC,
        REL,
        ALL
    }

    protected static Set<Rule> transformRules(ChaseInstance chaseInstance, Rule query, Set<Rule> rules,
            TRANSFORMATION transformation, Boolean doFastRelevanceAnalysis) throws FileNotFoundException {
        Set<Rule> transformedRules = new LinkedHashSet<Rule>();
        InputProcessor ruleCollector = new InputCollector(transformedRules, new LinkedHashSet<Atom>());
        RuleSubsumption ruleSubsumption = null;
        if (transformation.equals(TRANSFORMATION.MAGIC) || transformation.equals(TRANSFORMATION.ALL))
            ruleSubsumption = new RuleSubsumption(new RuleVerifier(ruleCollector));
        else
            ruleSubsumption = new RuleSubsumption(ruleCollector);
        OptimizedSIPs estimator = new OptimizedSIPs();
        InputProcessor chain = null;
        Map<Function, Integer> trueFunctions = new LinkedHashMap<>();
        String queryName = query.getHeadAtom(0).getPredicate().getName();

        Defun defun = null;
        if (chaseInstance.usesEquality())
            defun = new Defun(trueFunctions, Utilities.printer(chaseInstance, queryName, "P6"),
                    new Desingularization(Utilities.printer(chaseInstance, queryName, "P7"), ruleSubsumption));
        else
            defun = new Defun(trueFunctions, Utilities.printer(chaseInstance, queryName, "P6"), ruleSubsumption);

        if (transformation.equals(TRANSFORMATION.REL)) {
            chain = new Skolemization(true, true, true, new RuleRelevance(
                    query.getHeadAtom(0), new LinkedHashMap<>(),
                    chaseInstance.usesEquality(), doFastRelevanceAnalysis,
                    Utilities.printer(chaseInstance, queryName,
                            "P4"),
                    defun)) {
                protected int m_nextExistentialCounter = 0;

                protected int getNextExistentialCounter() {
                    return m_nextExistentialCounter++;
                }
            };
        } else if (transformation.equals(TRANSFORMATION.MAGIC)) {
            chain = new Skolemization(true, true, true,
                    new MagicProcessor(query.getHeadAtom(0),
                            estimator,
                            // new MaximumVariableCoverSIPs(),
                            // true,
                            new LinkedHashMap<>(),
                            Utilities.printer(chaseInstance, queryName,
                                    "P5"),
                            defun)) {

                protected int m_nextExistentialCounter = 0;

                protected int getNextExistentialCounter() {
                    return m_nextExistentialCounter++;
                }
            };
        } else if (transformation.equals(TRANSFORMATION.ALL)) {
            chain = new Skolemization(true, true, true,
                    new RuleRelevance(query.getHeadAtom(0), new LinkedHashMap<>(),
                            chaseInstance.usesEquality(), doFastRelevanceAnalysis,
                            Utilities.printer(chaseInstance, queryName,
                                    "P4"),
                            new MagicProcessor(query.getHeadAtom(0),
                                    estimator,
                                    new LinkedHashMap<>(),
                                    Utilities.printer(chaseInstance, queryName,
                                            "P5"),
                                    // new MaximumVariableCoverSIPs(),
                                    // true,
                                    defun))) {
                protected int m_nextExistentialCounter = 0;

                protected int getNextExistentialCounter() {
                    return m_nextExistentialCounter++;
                }
            };
        }
        if (chaseInstance.usesEquality())
            chain = new Singularization(query.getHeadAtom(0).getPredicate(),
                    Utilities.printer(chaseInstance, queryName, "P2"), chain);

        chain.startProcessing();
        for (Rule inputRule : rules)
            chain.processRule(inputRule);
        chain.endProcessing();

        CommonFormatProcessor P8 = Utilities.printer(chaseInstance, queryName, "P8");
        for (Rule inputRule : transformedRules)
            P8.processRule(inputRule);
        P8.endProcessing();

        return transformedRules;
    }
}
