package uk.ac.ox.cs.goalDriven.chaseBench;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.Utilities;
import uk.ac.ox.cs.goalDriven.QA.processors.Defun;
import uk.ac.ox.cs.goalDriven.QA.processors.Desingularization;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities;
import uk.ac.ox.cs.goalDriven.QA.utilities.CommonUtilities.PredicateInfo;
import uk.ac.ox.cs.goalDriven.RDFoxBridge.JRDFoxBridge;
import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.parser.CommonParser;
import uk.ac.ox.cs.goalDriven.processors.InputCollector;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class RunSOChase {

    public static void main(String[] args) throws Exception {
        ChaseInstance chaseInstance = new ChaseInstance();
        String scenarioNamePrefix = "";
        String scenarioName = "generated26";
        String scenarioNameSuffix = "";
        String dataVariantName = "1000";
        String queryFileName = "queries.txt";

        chaseInstance.initialize(scenarioNamePrefix, scenarioName, scenarioNameSuffix, dataVariantName, false);
        Set<Rule> queries = Utilities.loadQueries(chaseInstance, queryFileName);

        Set<Rule> rules = new LinkedHashSet<Rule>();
        Set<Atom> facts = new LinkedHashSet<Atom>();
        // Load the rules
        if (chaseInstance.m_stTGDsFile != null)
            CommonParser.parse(chaseInstance.m_stTGDsFile, new InputCollector(rules, facts));

        rules.addAll(queries);
        List<Predicate> queryPredicates = new ArrayList<>();
        for (Rule query : queries)
            queryPredicates.add(query.getHeadAtom(0).getPredicate());

        CommonFormatProcessor P6 = new CommonFormatProcessor(
                new PrintWriter(new File(chaseInstance.m_transformedProgramsRoot,
                        "P6.txt")));
        CommonFormatProcessor P7 = new CommonFormatProcessor(
                new PrintWriter(new File(chaseInstance.m_transformedProgramsRoot,
                        "P7.txt")));
        CommonFormatProcessor P8 = new CommonFormatProcessor(
                new PrintWriter(new File(chaseInstance.m_transformedProgramsRoot,
                        "P8.txt")));

        Map<Function, Integer> trueFunctions = CommonUtilities.getTrueFunctions(rules);
        Set<Rule> transformedRules = new LinkedHashSet<>();
        Defun defun = new Defun(trueFunctions,
                P6, new Desingularization(P7, new InputCollector(transformedRules, new LinkedHashSet<Atom>())));

        defun.startProcessing();
        for (Rule inputRule : rules)
            defun.processRule(inputRule);
        defun.endProcessing();

        Set<Rule> auxiliaryAxioms = Utilities
                .createAuxiliarySOAxiomsForBodyFunctions(defun.getBodyPredicates());
        transformedRules.addAll(auxiliaryAxioms);
        transformedRules.addAll(CommonUtilities.getDomainAxioms(rules));

        P8.startProcessing();
        for (Rule inputRule : transformedRules)
            P8.processRule(inputRule);
        P8.endProcessing();

        // Apply the chase using RDFox
        try (JRDFoxBridge jrdfoxBridge = new JRDFoxBridge(chaseInstance.usesEquality())) {
            Map<Predicate, PredicateInfo> predicateInfo = CommonUtilities.getPredicateInfo(transformedRules);
            InputProcessor inputProcessor = jrdfoxBridge.newInputProcessor();

            System.out.print("Loading the source instance... ");
            System.out.flush();
            long start = System.currentTimeMillis();
            Utilities.loadCSVResources(chaseInstance, inputProcessor, predicateInfo);
            long duration = System.currentTimeMillis() - start;
            System.out.println(duration + " ms");

            inputProcessor.startProcessing();
            for (Rule rule : transformedRules)
                inputProcessor.processRule(rule);
            inputProcessor.endProcessing();

            System.out.print("Applying " + +transformedRules.size() + " rules... ");
            System.out.flush();
            start = System.currentTimeMillis();
            jrdfoxBridge.computeMaterialization();
            duration = System.currentTimeMillis() - start;
            System.out.println(duration + " ms");

            String modelPath = scenarioName + "/" + "model" + "/" + dataVariantName;
            new File(modelPath).mkdirs();
            Integer[] entry = CommonUtilities.saveInstanceOnDisk(jrdfoxBridge,
                    modelPath, transformedRules, queryPredicates);
            System.out.println("Total facts " + entry[0] + "; total magic facts " + entry[1] + "; total D + F facts "
                    + entry[2] + "; useful facts " + (entry[0] - entry[1] - entry[2]));

        }

    }

}
