package uk.ac.ox.cs.goalDriven.QA.generator;

import static org.junit.Assert.assertTrue;

import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Function;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;

public class BenchmarkGenerator {

    private int N_Q = 120;
    private int N_rho = 100;
    private int N_p_arity = 2;
    private int N_f_arity = 1;
    private int N_F = 1;
    private int N_r = 2;
    // private int N_eq = 4;
    private int N_D = 1;
    private int N_edges = 4;
    private int N_T = 10;
    private String m_scenarioPath = "./generated26b";
    private int Total_EQ = 20;

    private final Map<Predicate, Integer> m_predicateCache = new LinkedHashMap<>();
    public final Map<Function, Integer> m_functionCache = new LinkedHashMap<>();
    private final Set<Constant> m_constantCache = new LinkedHashSet<>();
    private final Set<Term> m_termCache = new LinkedHashSet<>();

    private final List<Rule> m_rules = new ArrayList<>();
    private final List<Rule> m_queries = new ArrayList<>();
    static Random random = new Random();

    public static void main(String[] args) {

        BenchmarkGenerator generator = new BenchmarkGenerator();
        generator.generate();

    }

    public BenchmarkGenerator() {
    }

    public BenchmarkGenerator(int N_Q, int N_rho, int N_p_arity, int N_f_arity, int N_F, int N_r, int N_eq,
            int N_D, int N_T, String scenarioPath) {
        this.N_Q = N_Q;
        this.N_rho = N_rho;
        this.N_p_arity = N_p_arity;
        this.N_f_arity = N_f_arity;
        this.N_F = N_F;
        this.N_r = N_r;
        // this.N_eq = N_eq;
        this.N_D = N_D;
        this.N_T = N_T;
        this.m_scenarioPath = scenarioPath;
    }

    public List<Rule> generate() {
        Graph<Atom> Gamma = new Graph<Atom>();
        Set<Atom> T = new LinkedHashSet<>();
        Set<Atom> D = new LinkedHashSet<>();

        try {
            Files.createDirectories(Paths.get(m_scenarioPath));
        } catch (IOException e) {
            e.printStackTrace();
        }

        int total_EQ = 0;
        // Create queries with overlapping constants
        Set<Predicate> queryPredicates = new LinkedHashSet<>();
        for (int i = 0; i < N_Q; ++i) {
            int arity = random.nextInt(N_p_arity) + 1;
            Term[] terms = new Term[arity];
            for (int j = 0; j < arity; ++j)
                terms[j] = randomConstant();

            Predicate Qi = Predicate.create("Q" + i);
            queryPredicates.add(Qi);
            Atom queryAtom = Atom.create(Qi, terms);
            m_predicateCache.put(Qi, arity);
            T.add(queryAtom);
            D.add(queryAtom);
        }

        while (!T.isEmpty() && m_rules.size() < N_rho) {

            Iterator<Atom> iterator = T.iterator();
            Atom F = iterator.next();
            System.out.println(F);
            iterator.remove();
            D.add(F);

            int n_F = random.nextInt(N_F) + 1;
            for (int i = 0; i < n_F; ++i) {

                Rule r_bar;
                if (F.getPredicate().equals(Predicate.EQUALS))
                    r_bar = createGroundEqualityRule(F);
                else {
                    // @WARNING the number of relational and equality body atoms must respect some
                    // constraints
                    // int n_r = random.nextInt(N_r) + 1;
                    // int n_eq = random.nextInt(N_eq) + 1;
                    // n_r = N_r;
                    // n_eq = N_eq;

                    // Control the number of relational and equality atoms
                    int minimum_r = F.getNumberOfArguments();
                    int n_r = random.nextInt(N_r - minimum_r + 1) + minimum_r;
                    int maximum_eq = Math.floorDiv(n_r, 2) + 1;
                    int n_eq = 0;
                    if (maximum_eq > 0 && random.nextDouble() > 0.8 && total_EQ++ <= Total_EQ)
                        n_eq = 1;

                    AbstractMap.SimpleEntry<Rule, Rule> r = createRule(F, n_r, n_eq);
                    Rule rule = r.getKey();
                    if (queryPredicates.contains(rule.getHeadAtom(0).getPredicate()))
                        m_queries.add(rule);
                    else
                        m_rules.add(rule);
                    r_bar = r.getValue();
                }
                for (int bodyAtomIndex = 0; bodyAtomIndex < r_bar.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
                    Atom F_i = r_bar.getBodyAtom(bodyAtomIndex);
                    // if (!D.contains(F_i))
                    {
                        T.add(F_i);
                        D.add(F_i);
                        Gamma.addEdge(F, F_i);
                    }
                }
            }
        }

        System.out.println("\n\n\nGeneration of transfering axioms\n\n\n");

        T.addAll(Gamma.getSyncNodes());
        D.clear();
        D.addAll(Gamma.getSyncNodes());
        while (!T.isEmpty()) {
            Iterator<Atom> iterator = T.iterator();
            Atom F = iterator.next();
            System.out.println(F);
            iterator.remove();
            D.add(F);
            AbstractMap.SimpleEntry<Rule, Rule> r = createTransferRule(F);
            Rule rule = r.getKey();
            if (queryPredicates.contains(rule.getHeadAtom(0).getPredicate()))
                m_queries.add(rule);
            else
                m_rules.add(rule);
            Atom F_p = r.getValue().getBodyAtom(0);
            if (!D.contains(F_p)) {
                D.add(F_p);
                Boolean containsFT = false;
                for (int j = 0; j < F_p.getNumberOfArguments(); ++j) {
                    if (F_p.getArgument(j) instanceof FunctionTerm)
                        containsFT = true;
                }
                if (containsFT)
                    T.add(F_p);
                else {
                    // Add facts for the edb atoms
                    try {
                        String path = m_scenarioPath + "/data";
                        Files.createDirectories(Paths.get(path));
                        FileWriter writer = new FileWriter(path + "/" + F_p.getPredicate().getName() + ".csv");
                        int a = F_p.getNumberOfArguments();
                        Term[] t = new Term[a];
                        F_p.getArguments(t);
                        for (int j = 0; j < a; ++j)
                            assertTrue(t[j] instanceof Constant);

                        for (int i = 0; i < N_T; ++i) {
                            for (int j = 0; j < a; ++j) {
                                writer.write(((Constant) t[j]).getLexicalForm() + "_" + i);
                                if (j < a - 1)
                                    writer.write(",");
                            }
                            writer.write("\n");
                        }
                        writer.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        // Write the queries
        try {
            String path = m_scenarioPath + "/queries";
            Files.createDirectories(Paths.get(path));
            FileWriter writer = new FileWriter(path + "/queries.txt");
            for (Rule query : m_queries)
                writer.write(query + "\n");
            writer.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        // Write the rules
        try {
            String path = m_scenarioPath + "/dependencies";
            Files.createDirectories(Paths.get(path));
            FileWriter writer = new FileWriter(path + "/dependencies.txt");
            for (Rule rule : m_rules)
                writer.write(rule + "\n");
            writer.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return m_rules;
    }

    private Variable freshVariable() {
        char n = (char) (random.nextInt(26) + 'a');
        int i = 10 + random.nextInt(90);
        return Variable.create(("" + n + i).toUpperCase());
    }

    private Predicate freshRelationalPredicate() {
        while (true) {
            char n1 = (char) (random.nextInt(26) + 'a');
            char n2 = (char) (random.nextInt(26) + 'a');
            char n3 = (char) (random.nextInt(26) + 'a');
            char n4 = (char) (random.nextInt(26) + 'a');
            Predicate p = Predicate.create("" + n1 + n2 + n3 + n4);
            if (!m_predicateCache.containsKey(p)) {
                int arity = random.nextInt(N_p_arity) + 1;
                m_predicateCache.put(p, arity);
                return p;
            }
        }
    }

    private Predicate randomRelationalPredicate() {
        if (random.nextDouble() > 0.9 && m_predicateCache.size() > N_Q) {
            while (true) {
                Predicate p = getRandomSetElement(m_predicateCache.keySet());
                if (!p.getName().startsWith("Q") && !p.equals(Predicate.EQUALS))
                    return p;
            }
        } else
            return freshRelationalPredicate();
    }

    private Function randomFunction() {
        if (random.nextDouble() > 0.5 && m_functionCache.size() > 0)
            return getRandomSetElement(m_functionCache.keySet());
        else {
            while (true) {
                char n1 = (char) (random.nextInt(26) + 'a');
                char n2 = (char) (random.nextInt(26) + 'a');
                char n3 = (char) (random.nextInt(26) + 'a');
                Function p = Function.create("" + n1 + n2 + n3);
                if (!m_functionCache.containsKey(p)) {
                    int arity = random.nextInt(N_f_arity) + 1;
                    m_functionCache.put(p, arity);
                    return p;
                }
            }
        }
    }

    private Term randomGroundTerm(int max_depth) {
        if (max_depth == 0)
            return randomConstant();
        else {
            if (random.nextDouble() > 0.3 && m_termCache.size() > 0) {
                int l = 0;
                int MAX_TRIES = 1000;
                while (l < MAX_TRIES) {
                    Term t = getRandomSetElement(m_termCache);
                    if (t.depth() < max_depth)
                        return t;
                    ++l;
                }
                return randomConstant();
            } else {
                Function f = randomFunction();
                int arity = m_functionCache.get(f);
                Term terms[] = new Term[arity];
                for (int i = 0; i < arity; ++i)
                    terms[i] = randomGroundTerm(max_depth - 1);
                Term fterm = FunctionTerm.create(f, terms);
                m_termCache.add(fterm);
                return fterm;
            }
        }
    }

    private Constant randomConstant() {
        if (random.nextDouble() > 0.5 && m_constantCache.size() > 0)
            return getRandomSetElement(m_constantCache);
        else {
            while (true) {
                int n1 = random.nextInt(10000);
                Constant c = Constant.create("c" + n1);
                if (!m_constantCache.contains(c)) {
                    m_constantCache.add(c);
                    return c;
                }
            }
        }
    }

    public AbstractMap.SimpleEntry<Rule, Rule> createTransferRule(Atom F) {
        Predicate S = F.getPredicate();
        List<Term> s = new ArrayList<>();
        List<Term> s_p = new ArrayList<>();
        List<Term> t_p = new ArrayList<>();
        for (int i = 0; i < F.getNumberOfArguments(); ++i) {
            Term t_i = F.getArgument(i);
            if (t_i instanceof FunctionTerm) {
                int m = ((FunctionTerm) t_i).getNumberOfArguments();
                Function f = ((FunctionTerm) t_i).getFunction();
                Term[] x = new Term[m];
                for (int j = 0; j < m; ++j)
                    x[j] = freshVariable();
                Term fx = FunctionTerm.create(f, x);
                s.add(fx);
                s_p.addAll(Arrays.asList(x));
                t_p.addAll(Arrays.asList(((FunctionTerm) t_i).getArguments()));
            } else {
                Variable X = freshVariable();
                s.add(X);
                s_p.add(X);
                t_p.add(t_i);
            }
        }
        Predicate R = freshRelationalPredicate();
        Atom S_s = Atom.create(S, s.toArray(new Term[0]));
        Atom R_s_p = Atom.create(R, s_p.toArray(new Term[0]));
        Atom R_t_p = Atom.create(R, t_p.toArray(new Term[0]));
        Rule r = Rule.create(S_s, R_s_p);
        Rule r_bar = Rule.create(F, R_t_p);
        System.out.println(r);
        System.out.println(r_bar);
        System.out.println("-----------------------------------------------------------");
        return new AbstractMap.SimpleEntry<Rule, Rule>(r, r_bar);
    }

    public Rule createGroundEqualityRule(Atom groundEquality) {

        // If the equality is of the form f(\vec{t1}) = f(\vec{t2}), choose between FR,
        // T, S;
        // otherwise choose between T, EQ
        Boolean onFR = false;
        Term t1 = groundEquality.getArgument(0);
        Term t2 = groundEquality.getArgument(1);
        if (t1 instanceof FunctionTerm && t2 instanceof FunctionTerm
                && ((FunctionTerm) t1).getFunction().equals(((FunctionTerm) t2).getFunction()))
            onFR = true;
        EQUALITYAXIOMS equalityAxiom = EQUALITYAXIOMS.randomAxiom(onFR);
        Rule r_bar = null;
        // Transitivity
        if (equalityAxiom.equals(EQUALITYAXIOMS.T)) {
            // Create a new term to map to Z
            Term t3 = randomGroundTerm(N_D);
            r_bar = Rule.create(groundEquality, Atom.create(Predicate.EQUALS, t1, t3),
                    Atom.create(Predicate.EQUALS, t3, t2));
        }
        // Functional reflexivity
        else if (equalityAxiom.equals(EQUALITYAXIOMS.FR)) {
            Term[] v1 = ((FunctionTerm) t1).getArguments();
            Term[] v2 = ((FunctionTerm) t2).getArguments();
            List<Atom> bodyAtoms = new LinkedList<>();
            for (int i = 0; i < v1.length; ++i)
                bodyAtoms.add(Atom.create(Predicate.EQUALS, v1[i], v2[i]));
            r_bar = Rule.create(groundEquality, bodyAtoms.toArray(new Atom[0]));
        }
        // Symmetry
        else
            r_bar = Rule.create(groundEquality, Atom.create(Predicate.EQUALS, t2, t1));
        System.out.println(r_bar);
        System.out.println("-----------------------------------------------------------");
        return r_bar;
    }

    public AbstractMap.SimpleEntry<Rule, Rule> createRule(Atom F, int n_r, int n_eq) {

        int rows = 1 + n_r + n_eq;
        int columns = N_p_arity;

        Function Fs[][] = new Function[rows][columns];
        Term U[][][] = new Term[rows][columns][N_f_arity];
        Term U_bar[][][] = new Term[rows][columns][N_f_arity];
        Predicate[] P = new Predicate[1 + n_r + n_eq];
        // The edges are only used for cycle detection, as (ground) term propagation is
        // happening the time we add edges
        Set<Triple<Integer, Integer, Integer>> nodes_with_outgoing_edges = new LinkedHashSet<>();
        Set<Triple<Integer, Integer, Integer>> nodes_with_incoming_edges = new LinkedHashSet<>();

        P[0] = F.getPredicate();

        for (int j = 0; j < F.getNumberOfArguments(); ++j) {
            if (F.getArgument(j) instanceof FunctionTerm && random.nextDouble() > 0.5) {
                FunctionTerm term = ((FunctionTerm) F.getArgument(j));
                Function f = term.getFunction();
                int arity = term.getNumberOfArguments();
                Fs[0][j] = f;
                for (int k = 0; k < arity; ++k) {
                    U[0][j][k] = freshVariable();
                    U_bar[0][j][k] = term.getArgument(k);
                }
            } else {
                Fs[0][j] = null;
                U[0][j][0] = freshVariable();
                U_bar[0][j][0] = F.getArgument(j);
            }
        }

        for (int i = 1; i < n_r + 1; ++i) {
            P[i] = randomRelationalPredicate();
            for (int j = 0; j < m_predicateCache.get(P[i]); ++j) {
                Fs[i][j] = null;

                // Initialize with fresh variables and random ground terms.
                // This is done to simplify the next step as (ground) term propagation is
                // happening the same time we add edges
                U[i][j][0] = freshVariable();
                U_bar[i][j][0] = randomGroundTerm(N_D);
            }
        }

        for (int i = n_r + 1; i < rows; ++i) {
            P[i] = Predicate.EQUALS;
            m_predicateCache.put(Predicate.EQUALS, 2);
            for (int j = 0; j < 2; ++j) {
                if (random.nextDouble() > 0.25)
                    Fs[i][j] = null;
                else
                    Fs[i][j] = randomFunction();
            }
        }

        // Add edges
        // • graph G is acyclic,
        // • each head position vertex has zero incoming edges and at least one outgoing
        // edge to a relational body position vertex,
        // • each relational body position vertex has at most one incoming edge, and
        // • each equality body position vertex has one incoming edge and zero outgoing
        // edges.

        // Add edges from head positions to a different body relational position
        // We check that each node has at most one incoming edge
        // Term propagation is happening at the same time
        // Choose edges positions between 1 and n_r + 1
        for (int j = 0; j < m_predicateCache.get(P[0]); ++j) {
            Function f = Fs[0][j];
            int arity = 0;
            if (f != null)
                arity = m_functionCache.get(f);
            // @WARNING equalities may increase the maximum allowable depth of terms by one.
            for (int k = 0; k < Math.max(arity, 1); ++k) {
                Triple<Integer, Integer, Integer> destination = randomRelationalDestinationPosition(P, Fs, 1,
                        n_r + 1, nodes_with_outgoing_edges, nodes_with_incoming_edges);
                if (destination == null)
                    throw new RuntimeException("Error while adding edges in position graph");
                else {
                    int i2 = destination.getFirst();
                    int j2 = destination.getSecond();
                    int k2 = destination.getThird();
                    nodes_with_outgoing_edges.add(new Triple<Integer, Integer, Integer>(0, j, k));
                    nodes_with_incoming_edges.add(new Triple<Integer, Integer, Integer>(i2, j2, k2));
                    U[i2][j2][k2] = U[0][j][k];
                    U_bar[i2][j2][k2] = U_bar[0][j][k];
                }
            }
        }

        // Add edges from body relational positions to different body relational
        // positions
        // The chosen destination node must have no incoming and no outgoing edges to
        // satisfy the second and the third conditions
        // Term propagation is happening at the same time
        int n_e = 0;
        while (n_e < N_edges) {
            Triple<Integer, Integer, Integer> source = randomRelationalSourcePosition(P, Fs, 1, n_r + 1);
            Triple<Integer, Integer, Integer> destination = randomRelationalDestinationPosition(P, Fs, 1, n_r + 1,
                    nodes_with_outgoing_edges, nodes_with_incoming_edges);
            if (destination == null)
                break;
            else {
                int i = source.getFirst();
                int j = source.getSecond();
                assertTrue(source.getThird() == 0);
                int i2 = destination.getFirst();
                int j2 = destination.getSecond();
                assertTrue(destination.getThird() == 0);
                nodes_with_outgoing_edges.add(source);
                nodes_with_incoming_edges.add(destination);
                U[i2][j2][0] = U[i][j][0];
                U_bar[i2][j2][0] = U_bar[i][j][0];
            }
            ++n_e;
        }

        // Add edges from body relational positions to different body equality positions
        // (we iterate over body equality atoms)
        // Term propagation is happening at the same time
        // Here, we simply choose relational positions-- no other check
        for (int i2 = 1 + n_r; i2 < 1 + n_r + n_eq; ++i2) {
            for (int j2 = 0; j2 < 2; ++j2) {
                int f_arity = 0;
                if (Fs[i2][j2] != null)
                    f_arity = m_functionCache.get(Fs[i2][j2]);
                for (int k2 = 0; k2 < Math.max(f_arity, 1); ++k2) {
                    Triple<Integer, Integer, Integer> source = randomRelationalSourcePosition(P, Fs, 1, n_r + 1);
                    int i = source.getFirst();
                    int j = source.getSecond();
                    assertTrue(source.getThird() == 0);
                    U[i2][j2][k2] = U[i][j][0];
                    U_bar[i2][j2][k2] = U_bar[i][j][0];
                }
            }
        }

        // Create the rules
        Atom groundHead = null;
        Atom nongroundHead = null;
        Atom[] groundAtoms = new Atom[P.length - 1];
        Atom[] nongroundAtoms = new Atom[P.length - 1];

        for (int i = 0; i < P.length; ++i) {
            Predicate p = P[i];
            int p_arity = m_predicateCache.get(p);
            Term nongroundTerms[] = new Term[p_arity];
            Term groundTerms[] = new Term[p_arity];
            for (int j = 0; j < p_arity; ++j) {
                Term gt;
                Term ngt;
                if (Fs[i][j] == null) {
                    gt = U_bar[i][j][0];
                    ngt = U[i][j][0];
                } else {
                    Function f = Fs[i][j];
                    int f_arity = m_functionCache.get(Fs[i][j]);
                    Term terms2[] = new Term[f_arity];
                    Term terms1[] = new Term[f_arity];
                    for (int k = 0; k < f_arity; ++k) {
                        terms2[k] = U_bar[i][j][k];
                        terms1[k] = U[i][j][k];
                    }
                    gt = FunctionTerm.create(f, terms2);
                    ngt = FunctionTerm.create(f, terms1);
                }
                groundTerms[j] = gt;
                nongroundTerms[j] = ngt;
            }
            if (i == 0) {
                groundHead = Atom.create(P[i], groundTerms);
                nongroundHead = Atom.create(P[i], nongroundTerms);
            } else {
                groundAtoms[i - 1] = Atom.create(P[i], groundTerms);
                nongroundAtoms[i - 1] = Atom.create(P[i], nongroundTerms);
            }
        }

        // After creating the rule, it may have cartesian products.
        // To overcome these cases, we create a new atom with a fresh predicate.
        // The arguments of this atoms are one variable from each connected component
        // We also update the corresponding ground rule.
        // Create a substitution map from variables to ground terms.
        Map<Variable, Term> substitution = new LinkedHashMap<>();
        Graph<Variable> variableGraph = new Graph<>();
        for (int nonGroundAtomIndex = 0; nonGroundAtomIndex < nongroundAtoms.length; ++nonGroundAtomIndex) {
            Atom atom = nongroundAtoms[nonGroundAtomIndex];
            if (!atom.getPredicate().equals(Predicate.EQUALS)) {
                Atom groundAtom = groundAtoms[nonGroundAtomIndex];
                variableGraph.addVertex((Variable) atom.getArgument(0));
                // As all variables occurring in the head or equalities also occur in a
                // relational body atom as a top-level term
                // we collect all variables in this step;
                substitution.put((Variable) atom.getArgument(0), groundAtom.getArgument(0));
                for (int argumentIndex = 1; argumentIndex < atom.getNumberOfArguments(); ++argumentIndex) {
                    substitution.put((Variable) atom.getArgument(argumentIndex), groundAtom.getArgument(argumentIndex));
                    variableGraph.addEdge((Variable) atom.getArgument(argumentIndex - 1),
                            (Variable) atom.getArgument(argumentIndex));
                    variableGraph.addEdge((Variable) atom.getArgument(argumentIndex),
                            (Variable) atom.getArgument(argumentIndex - 1));
                }
            }
        }
        Rule r = null;
        Rule r_bar = null;
        // Find the connected components of the rule body, seen as a graph where there
        // exists an edge between two variables if they co-occur in the same atom.
        // If there is a single connected component do nothing.
        // Else follow the above steps.
        Set<Set<Variable>> components = variableGraph.connectedComponents();
        if (components.size() > 1) {
            Variable[] guardVariables = new Variable[components.size()];
            Term[] guardTerms = new Term[components.size()];
            int i = 0;
            for (Set<Variable> componenent : components) {
                Variable v = componenent.iterator().next();
                guardVariables[i] = v;
                guardTerms[i] = substitution.get(v);
                ++i;
            }
            Predicate p = freshRelationalPredicate();
            m_predicateCache.put(p, components.size());
            Atom nongroundGuard = Atom.create(p, guardVariables);
            Atom groundGuard = Atom.create(p, guardTerms);
            Atom[] nongroundAtoms2 = new Atom[nongroundAtoms.length + 1];
            Atom[] groundAtoms2 = new Atom[groundAtoms.length + 1];
            System.arraycopy(nongroundAtoms, 0, nongroundAtoms2, 0, nongroundAtoms.length);
            System.arraycopy(groundAtoms, 0, groundAtoms2, 0, groundAtoms.length);
            nongroundAtoms2[nongroundAtoms.length] = nongroundGuard;
            groundAtoms2[groundAtoms.length] = groundGuard;
            r = Rule.create(nongroundHead, nongroundAtoms2);
            r_bar = Rule.create(groundHead, groundAtoms2);

        } else {
            r = Rule.create(nongroundHead, nongroundAtoms);
            r_bar = Rule.create(groundHead, groundAtoms);
        }

        System.out.println(r);
        System.out.println(r_bar);
        System.out.println("-----------------------------------------------------------");
        return new AbstractMap.SimpleEntry<Rule, Rule>(r, r_bar);
    }

    private Triple<Integer, Integer, Integer> randomRelationalSourcePosition(Predicate[] P, Function Fs[][],
            int min_p_index, int max_p_index) {
        int i = random.nextInt(max_p_index - min_p_index) + min_p_index;
        int p_arity = m_predicateCache.get(P[i]);
        int j = random.nextInt(p_arity - 0);
        int k = 0;
        if (Fs[i][j] != null) {
            int f_arity = m_functionCache.get(Fs[i][j]);
            k = random.nextInt(f_arity);
        }
        return new Triple<Integer, Integer, Integer>(i, j, k);
    }

    private Triple<Integer, Integer, Integer> randomRelationalDestinationPosition(Predicate[] P,
            Function Fs[][], int min_p_index, int max_p_index,
            Set<Triple<Integer, Integer, Integer>> nodes_with_outgoing_edges,
            Set<Triple<Integer, Integer, Integer>> nodes_with_incoming_edges) {

        int l = 0;
        int MAX_TRIES = 1000;
        while (l < MAX_TRIES) {
            int i = random.nextInt(max_p_index - min_p_index) + min_p_index;
            int p_arity = m_predicateCache.get(P[i]);
            int j = random.nextInt(p_arity - 0);
            int k = 0;
            if (Fs[i][j] != null) {
                int f_arity = m_functionCache.get(Fs[i][j]);
                k = random.nextInt(f_arity);
            }
            Triple<Integer, Integer, Integer> destination = new Triple<Integer, Integer, Integer>(i, j, k);
            if (!nodes_with_outgoing_edges.contains(destination) && !nodes_with_incoming_edges.contains(destination))
                return destination;
            ++l;
        }
        return null;
    }

    enum EQUALITYAXIOMS {
        T,
        S,
        FR;

        public static EQUALITYAXIOMS randomAxiom(Boolean onFR) {
            EQUALITYAXIOMS[] axioms = values();
            if (onFR)
                return axioms[random.nextInt(3)];
            else
                return axioms[random.nextInt(2)];
        }
    }

    class Graph<T> {

        private Map<T, List<T>> map = new HashMap<>();

        public void addVertex(T s) {
            if (!map.containsKey(s))
                map.put(s, new LinkedList<T>());
        }

        public void addEdge(T source, T destination) {

            if (!map.containsKey(source))
                addVertex(source);

            if (!map.containsKey(destination))
                addVertex(destination);

            map.get(source).add(destination);
        }

        public Set<T> getSyncNodes() {
            Set<T> nodes = new LinkedHashSet<>();
            for (Map.Entry<T, List<T>> e : map.entrySet()) {
                if (e.getValue().isEmpty())
                    nodes.add(e.getKey());
            }
            return nodes;
        }

        Set<T> DFSUtil(T v, Map<T, Boolean> visited) {
            Set<T> reachable = new LinkedHashSet<>();
            reachable.add(v);
            visited.put(v, true);
            for (T x : map.get(v)) {
                if (!visited.get(x))
                    reachable.addAll(DFSUtil(x, visited));
            }
            return reachable;
        }

        Set<Set<T>> connectedComponents() {
            Set<Set<T>> connectedComponents = new LinkedHashSet<>();
            Map<T, Boolean> visited = new LinkedHashMap<>();
            for (T key : this.map.keySet())
                visited.put(key, false);
            for (T v : this.map.keySet()) {
                if (!visited.get(v))
                    connectedComponents.add(DFSUtil(v, visited));
            }
            return connectedComponents;
        }

    }

    public static <T> T getRandomSetElement(Set<T> set) {
        if (set == null || set.isEmpty()) {
            throw new IllegalArgumentException("The Set cannot be empty.");
        }
        int randomIndex = random.nextInt(set.size());
        int i = 0;
        for (T element : set) {
            if (i == randomIndex) {
                return element;
            }
            i++;
        }
        throw new IllegalStateException("Something went wrong while picking a random element.");
    }

    public class Triple<T, U, V> {

        private final T first;
        private final U second;
        private final V third;

        public Triple(T first, U second, V third) {
            this.first = first;
            this.second = second;
            this.third = third;
        }

        public T getFirst() {
            return first;
        }

        public U getSecond() {
            return second;
        }

        public V getThird() {
            return third;
        }

        @SuppressWarnings("rawtypes")
        @Override
        public boolean equals(Object o) {
            if (this == o)
                return true;
            if (o == null)
                return false;
            return (o instanceof Triple) && this.first == (((Triple) o).getFirst())
                    && this.second == (((Triple) o).getSecond()) && this.third.equals(((Triple) o).getThird());
        }

        @Override
        public int hashCode() {
            return 31 + this.first.hashCode() * 7 + this.second.hashCode() * 9 + this.third.hashCode() * 11;
        }

        @Override
        public String toString() {
            return "<" + this.first.toString() + "," + this.second.toString() + "," + this.third.toString() + ">";
        }
    }

}
