package uk.ac.ox.cs.goalDriven.QA.magic;

import java.util.Arrays;

import uk.ac.ox.cs.goalDriven.model.InterningManager;
import uk.ac.ox.cs.goalDriven.model.Predicate;

public class Binding {
    protected final Predicate m_predicate;
    protected final Boolean[] m_adornment;

    private Binding(Predicate predicate, Boolean[] adornment) {
        m_predicate = predicate;
        m_adornment = adornment;
    }

    public boolean subsumes(Binding pattern) {
        if (m_predicate.equals(pattern.m_predicate)
                && !Arrays.equals(m_adornment, pattern.m_adornment)) {
            for (int adornmentIndex = 0; adornmentIndex < m_adornment.length; ++adornmentIndex) {
                if (m_adornment[adornmentIndex] == true && pattern.m_adornment[adornmentIndex] == false)
                    return false;
            }
            return true;
        }
        return false;
    }

    protected static final InterningManager<Binding> s_interningManager = new InterningManager<Binding>() {
        protected boolean equal(Binding object1, Binding object2) {
            if (!object1.m_predicate.equals(object2.m_predicate))
                return false;
            for (int index = object1.m_adornment.length - 1; index >= 0; --index)
                if (!object1.m_adornment[index].equals(object2.m_adornment[index]))
                    return false;
            return true;
        }

        protected int getHashCode(Binding object) {
            int hashCode = object.m_predicate.hashCode();
            for (int index = object.m_adornment.length - 1; index >= 0; --index)
                hashCode = hashCode * 7 + object.m_adornment[index].hashCode();
            return hashCode;
        }
    };

    public static Binding create(Predicate predicate, Boolean[] adornment) {
        return s_interningManager.intern(new Binding(predicate, adornment));
    }

    public Predicate getPredicate() {
        return m_predicate;
    }

    public Boolean[] getAdornment() {
        return m_adornment;
    }

}