package uk.ac.ox.cs.goalDriven.QA.magic.sips;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.QA.magic.Binding;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Variable;

public abstract class SIPs {
    public abstract Atom[] getBodyOrdering(Rule rule, Set<Variable> variables);

    public Atom[] getBodyOrderingLegacy(Rule rule, Boolean[] headAdornment) {
        Set<Variable> variables = new LinkedHashSet<>();
        Atom headAtom = rule.getHeadAtom(0);
        for (int index = 0; index < headAdornment.length; ++index) {
            if (headAdornment[index])
                headAtom.getFreeVariables(variables);
        }
        return getBodyOrdering(rule, variables);
    }

    public Entry<Atom[], Boolean[][]> sips(Rule rule, Set<Variable> variables,
            Map<Binding, Binding> adornmentReplacements) {
        Atom[] ordering = getBodyOrdering(rule, variables);
        Boolean[][] adornments = new Boolean[ordering.length][];
        Set<Variable> observedVariables = new LinkedHashSet<>();
        observedVariables.addAll(variables);
        for (int atomIndex = 0; atomIndex < ordering.length; ++atomIndex) {
            // Cover as many arguments as possible unless it is an equality atom where only
            // one of the two arguments should be bound
            Atom orderedAtom = ordering[atomIndex];
            Boolean[] adornment = new Boolean[orderedAtom.getNumberOfArguments()];
            for (int argumentIndex = 0; argumentIndex < orderedAtom.getNumberOfArguments(); ++argumentIndex) {
                List<Variable> freeVariables = new ArrayList<>();
                orderedAtom.getArgument(argumentIndex).getFreeVariables(freeVariables);
                if (freeVariables.size() == 0 || observedVariables.containsAll(freeVariables))
                    adornment[argumentIndex] = true;
                else
                    adornment[argumentIndex] = false;
                // In the case of equalities only one position is bound, so as soon as a
                // position is bound, we break
                if (orderedAtom.getPredicate().equals(Predicate.EQUALS) && adornment[argumentIndex] == true) {
                    if (argumentIndex == 0)
                        adornment[1] = false;
                    break;
                }
            }
            orderedAtom.getFreeVariables(observedVariables);
            Binding b = Binding.create(orderedAtom.getPredicate(), adornment);
            if (adornmentReplacements.containsKey(b))
                adornments[atomIndex] = adornmentReplacements.get(b).getAdornment();
            else
                adornments[atomIndex] = adornment;
        }

        return new AbstractMap.SimpleEntry<Atom[], Boolean[][]>(ordering, adornments);
    }
}
