package uk.ac.ox.cs.goalDriven.QA.processors;

import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class Desingularization implements InputProcessor {
    protected final InputProcessor m_inputProcessor;
    protected final CommonFormatProcessor m_writter;

    public Desingularization(CommonFormatProcessor writter, InputProcessor inputProcessor) {
        m_inputProcessor = inputProcessor;
        m_writter = writter;
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
        m_writter.startProcessing();

    }

    public void processRule(Rule rule) {
        assertTrue(rule.getNumberOfHeadAtoms() == 1);
        Set<Integer> positionsOfNonTrivialEqualities = new LinkedHashSet<>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            Atom bodyAtom = rule.getBodyAtom(bodyAtomIndex);
            if (!(Predicate.EQUALS.equals(bodyAtom.getPredicate())
                    && (bodyAtom.getArgument(0) instanceof Variable && bodyAtom.getArgument(1) instanceof Constant
                            || bodyAtom.getArgument(1) instanceof Variable
                                    && bodyAtom.getArgument(0) instanceof Constant
                            || bodyAtom.getArgument(0) instanceof Variable
                                    && bodyAtom.getArgument(1) instanceof Variable))) {
                positionsOfNonTrivialEqualities.add(bodyAtomIndex);
            }
        }
        m_inputProcessor.processRule(processRule(rule, positionsOfNonTrivialEqualities));
        m_writter.processRule(processRule(rule, positionsOfNonTrivialEqualities));
    }

    public static Rule processRule(Rule rule, Set<Integer> positionsOfNonTrivialEqualities) {
        assertTrue(rule.getNumberOfHeadAtoms() == 1);
        Atom head = rule.getHeadAtom(0);
        List<Atom> trivialEqualities = new ArrayList<>();
        List<Atom> nonTrivialEqualities = new ArrayList<>();
        // Get the trivial equalities
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            if (positionsOfNonTrivialEqualities.contains(bodyAtomIndex)
                    || !rule.getBodyAtom(bodyAtomIndex).getPredicate().equals(Predicate.EQUALS))
                nonTrivialEqualities.add(rule.getBodyAtom(bodyAtomIndex));
            else
                trivialEqualities.add(rule.getBodyAtom(bodyAtomIndex));
        }

        for (int i = 0; i < trivialEqualities.size(); ++i) {
            Map<Term, Term> substitution = new LinkedHashMap<Term, Term>();
            Atom equality = trivialEqualities.get(i);
            if (equality.getArgument(0) instanceof Variable)
                substitution.put(equality.getArgument(0), equality.getArgument(1));
            else
                substitution.put(equality.getArgument(1), equality.getArgument(0));

            // Apply the substitution to the remaining equalities
            for (int j = i + 1; j < trivialEqualities.size(); ++j)
                trivialEqualities.set(j, trivialEqualities.get(j).applyMapping(substitution));

            // Apply the substitution to the remaining equalities
            for (int j = 0; j < nonTrivialEqualities.size(); ++j)
                nonTrivialEqualities.set(j, nonTrivialEqualities.get(j).applyMapping(substitution));

            head = head.applyMapping(substitution);
        }
        Set<Atom> nonTrivialEqualitiesDeduplicate = new LinkedHashSet<>(nonTrivialEqualities);
        return Rule.create(head,
                nonTrivialEqualitiesDeduplicate.toArray(new Atom[nonTrivialEqualitiesDeduplicate.size()]));
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
        m_writter.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {
        m_inputProcessor.endProcessing();
        m_writter.endProcessing();
    }

}
