package uk.ac.ox.cs.goalDriven.QA.processors;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.FunctionTerm;
import uk.ac.ox.cs.goalDriven.model.LabeledNull;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class ReplaceFunctionTermsWithNulls implements InputProcessor {
    protected final Map<Atom, Atom> m_atomsMap;
    protected final InputProcessor m_inputProcessor;

    public ReplaceFunctionTermsWithNulls(InputProcessor inputProcessor) {
        m_inputProcessor = inputProcessor;
        m_atomsMap = new LinkedHashMap<>();
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
    }

    public void processRule(Rule rule) {
        Atom[] newHead = new Atom[rule.getNumberOfHeadAtoms()];
        for (int headIndex = 0; headIndex < rule.getNumberOfHeadAtoms(); ++headIndex) {
            Atom headAtom = rule.getHeadAtom(headIndex);
            Term[] newTerms = new Term[headAtom.getNumberOfArguments()];
            boolean replaceTerms = false;
            for (int argumentIndex = 0; argumentIndex < headAtom.getNumberOfArguments(); ++argumentIndex) {
                Term term = headAtom.getArgument(argumentIndex);
                if (term instanceof FunctionTerm) {
                    replaceTerms = true;
                    newTerms[argumentIndex] = LabeledNull.create(((FunctionTerm)term).getFunction().getName());
                } else
                    newTerms[argumentIndex] = term;
            }
            if (replaceTerms)
                m_atomsMap.put(Atom.create(headAtom.getPredicate(), newTerms), headAtom);
            newHead[headIndex] = Atom.create(headAtom.getPredicate(), newTerms);
        }
        m_inputProcessor.processRule(Rule.create(newHead, rule.getBodyAtoms()));
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {
        m_inputProcessor.endProcessing();
    }

    public Map<Atom, Atom> getAtomsMap() {
        return m_atomsMap;
    }
}
