package uk.ac.ox.cs.goalDriven.QA.processors;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

public class RuleSimplification implements InputProcessor {
    protected final InputProcessor m_inputProcessor;
    protected final Set<Rule> m_rules;
    protected final Set<Rule> m_units;

    public RuleSimplification(InputProcessor inputProcessor) {
        m_inputProcessor = inputProcessor;
        m_rules = new LinkedHashSet<>();
        m_units = new LinkedHashSet<>();
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
    }

    public void processRule(Rule rule) {
        if(rule.getNumberOfBodyAtoms() == 0)
            m_units.add(rule);
        m_rules.add(rule);
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {
        Rule[] rules = m_rules.toArray(new Rule[m_rules.size()]);
        Set<Rule> units = new LinkedHashSet<>();
        units.addAll(m_units);
        Set<Rule> pastUnits = new LinkedHashSet<>();
        pastUnits.addAll(m_units);
        
        Iterator<Rule> unitIterator = units.iterator();
        while(unitIterator.hasNext()) {
            Rule unit = unitIterator.next();
            unitIterator.remove();
            for(int ruleIndex = 0; ruleIndex < rules.length; ++ruleIndex) {
                Rule rule = rules[ruleIndex];
                if(!unit.equals(rule)) {
                    if(Arrays.asList(rule.getBodyAtoms()).contains(unit.getHeadAtom(0))) {
                        Atom[] newBody = new Atom[rule.getNumberOfBodyAtoms()-1];
                        int newBodyIndex = 0;
                        for(Atom atom:rule.getBodyAtoms()) {
                            if(!atom.equals(unit.getHeadAtom(0)))
                                newBody[newBodyIndex++] = atom;
                        }
                        rules[ruleIndex] = Rule.create(rule.getHeadAtom(0), newBody);
                        if(rules[ruleIndex].getNumberOfBodyAtoms() == 0 && !pastUnits.contains(rules[ruleIndex])) { 
                            units.add(rules[ruleIndex]);
                            pastUnits.add(rules[ruleIndex]);
                            unitIterator = units.iterator();
                        }
                    }
                }
            }
        }

        for(int ruleIndex1 = 0; ruleIndex1 < rules.length; ++ruleIndex1) {
            if(rules[ruleIndex1] != null) {
                Rule rule1 = rules[ruleIndex1];
                for(int ruleIndex2 = 0; ruleIndex2 < rules.length; ++ruleIndex2) {
                    if(rules[ruleIndex2] != null && !rules[ruleIndex1].equals(rules[ruleIndex2])) {
                        Rule rule2 = rules[ruleIndex2];
                        if(Arrays.equals(rule1.getHeadAtoms(), rule2.getHeadAtoms())) {
                            if(Arrays.asList(rule2.getBodyAtoms()).containsAll(Arrays.asList(rule1.getBodyAtoms()))) 
                                rules[ruleIndex2] = null;
                        }
                    }
                }
            }
        }

        for(Rule rule:rules) {
            if(rule != null) 
                m_inputProcessor.processRule(rule);
        }
        m_inputProcessor.endProcessing();
    }
}
