package uk.ac.ox.cs.goalDriven.QA.processors;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import uk.ac.ox.cs.goalDriven.format.CommonFormatProcessor;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Constant;
import uk.ac.ox.cs.goalDriven.model.Predicate;
import uk.ac.ox.cs.goalDriven.model.Rule;
import uk.ac.ox.cs.goalDriven.model.Term;
import uk.ac.ox.cs.goalDriven.model.Variable;
import uk.ac.ox.cs.goalDriven.processors.InputProcessor;

import java.util.Set;

public class Singularization implements InputProcessor {
    protected final InputProcessor m_inputProcessor;
    protected final Predicate m_queryPredicate;
    protected final CommonFormatProcessor m_writter;

    public Singularization(Predicate queryPredicate, CommonFormatProcessor writter, InputProcessor inputProcessor) {
        m_queryPredicate = queryPredicate;
        m_inputProcessor = inputProcessor;
        m_writter = writter;
    }

    public void startProcessing() {
        m_inputProcessor.startProcessing();
        m_writter.startProcessing();
    }

    public void processRule(Rule rule) {
        int variableCounter = 0;
        Map<Variable, Set<Entry<Integer, Integer>>> occurences = findOccurences(rule);
        List<Atom> newBody = new ArrayList<Atom>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            Atom bodyAtom = rule.getBodyAtom(bodyAtomIndex);
            if (bodyAtom.getPredicate() != Predicate.EQUALS) {
                List<Atom> newEqualities = new ArrayList<Atom>();
                Term[] newArguments = bodyAtom.getArguments();
                for (int argumentIndex = 0; argumentIndex < newArguments.length; ++argumentIndex) {
                    Term argument = newArguments[argumentIndex];
                    if (argument instanceof Constant) {
                        Variable newVariable = Variable.create("sg" + variableCounter++);
                        newArguments[argumentIndex] = newVariable;
                        Term[] eqArguments = { argument, newVariable };
                        newEqualities.add(Atom.create(Predicate.EQUALS, eqArguments));
                    } else if (argument instanceof Variable) {
                        Set<Entry<Integer, Integer>> s = occurences.get(argument);
                        if (s.size() > 1) {
                            s.remove(new AbstractMap.SimpleEntry<Integer, Integer>(bodyAtomIndex, argumentIndex));
                            occurences.put((Variable) argument, s);
                            Variable newVariable = Variable.create("sg" + variableCounter++);
                            newArguments[argumentIndex] = newVariable;
                            Term[] eqArguments = { argument, newVariable };
                            newEqualities.add(Atom.create(Predicate.EQUALS, eqArguments));
                        }
                    }
                }
                newBody.add(Atom.create(bodyAtom.getPredicate(), newArguments));
                newBody.addAll(newEqualities);
            } else
                newBody.add(bodyAtom);
        }

        Rule singularizedRule = null;
        if (m_queryPredicate != null && rule.getHeadAtom(0).getPredicate().equals(m_queryPredicate)) {
            Atom headAtom = rule.getHeadAtom(0);
            Term[] newHeadvariables = new Term[headAtom.getNumberOfArguments()];
            for (int argumentIndex = 0; argumentIndex < headAtom.getNumberOfArguments(); ++argumentIndex) {
                Variable variable = (Variable) headAtom.getArgument(argumentIndex);
                Variable newVariable = Variable.create("sg" + variableCounter++);
                Term[] eqArguments = { variable, newVariable };
                newBody.add(Atom.create(Predicate.EQUALS, eqArguments));
                newHeadvariables[argumentIndex] = newVariable;
            }
            singularizedRule = Rule.create(Atom.create(rule.getHeadAtom(0).getPredicate(), newHeadvariables),
                    newBody.toArray(new Atom[newBody.size()]));
        } else
            singularizedRule = Rule.create(rule.getHeadAtoms(), newBody.toArray(new Atom[newBody.size()]));

        m_inputProcessor.processRule(singularizedRule);
        m_writter.processRule(singularizedRule);
    }

    private Map<Variable, Set<Entry<Integer, Integer>>> findOccurences(Rule rule) {
        // Mark the first occurrence of each join variable
        Map<Variable, Set<Entry<Integer, Integer>>> occurences = new LinkedHashMap<>();
        for (int bodyAtomIndex = 0; bodyAtomIndex < rule.getNumberOfBodyAtoms(); ++bodyAtomIndex) {
            Atom bodyAtom = rule.getBodyAtom(bodyAtomIndex);
            if (bodyAtom.getPredicate() != Predicate.EQUALS) {
                for (int argumentIndex = 0; argumentIndex < bodyAtom.getNumberOfArguments(); ++argumentIndex) {
                    Term argument = bodyAtom.getArgument(argumentIndex);
                    if (argument instanceof Variable) {
                        Set<Entry<Integer, Integer>> s = occurences.get((Variable) argument);
                        if (s == null)
                            s = new LinkedHashSet<>();
                        s.add(new AbstractMap.SimpleEntry<Integer, Integer>(bodyAtomIndex, argumentIndex));
                        occurences.put((Variable) argument, s);
                    }
                }
            }
        }
        return occurences;
    }

    public void processFact(Predicate predicate, List<String> argumentLexicalForms) {
        m_inputProcessor.processFact(predicate, argumentLexicalForms);
        m_writter.processFact(predicate, argumentLexicalForms);
    }

    public void endProcessing() {
        m_inputProcessor.endProcessing();
        m_writter.endProcessing();
    }
}
