package uk.ac.ox.cs.goalDriven.RDFoxBridge;

import java.util.List;
import java.util.ArrayList;
import java.util.function.Function;

import uk.ac.ox.cs.JRDFox.JRDFoxException;
import uk.ac.ox.cs.JRDFox.store.Resource;
import uk.ac.ox.cs.JRDFox.store.TupleIterator;
import uk.ac.ox.cs.goalDriven.model.Atom;
import uk.ac.ox.cs.goalDriven.model.Variable;

public class JRDFoxBridgeUtilities {
    public static List<List<String>> evaluateQuery(JRDFoxBridge jrdfoxBridge, List<Variable> answerVariables,
            List<Atom> queryAtoms, Function<Resource, ?> resourceValueFunction, boolean onlyCertainAnswers,
            boolean printAnswers)
            throws JRDFoxException {
        List<List<String>> answers = new ArrayList<>();

        if (printAnswers) {
            boolean first = true;
            System.out.print('(');
            for (Variable variable : answerVariables) {
                if (first)
                    first = false;
                else
                    System.out.print(", ");
                System.out.print(variable.toString());
            }
            System.out.print(") :- ");
            first = true;
            for (Atom atom : queryAtoms) {
                if (first)
                    first = false;
                else
                    System.out.print(", ");
                System.out.print(atom.toString());
            }
            System.out.println(" .");
        }

        try (TupleIterator tupleIterator = jrdfoxBridge.createTupleIterator(true, answerVariables, queryAtoms,
                onlyCertainAnswers)) {
            int arity = tupleIterator.getArity();
            if (printAnswers) {
                System.out.println("--------------------------------------------------------------------------------");
                for (int index = 0; index < arity; ++index) {
                    if (index != 0)
                        System.out.print(" ");
                    System.out.print(answerVariables.get(index));
                }
                System.out.println();
                System.out.println("--------------------------------------------------------------------------------");

            }
            int numberOfTuples = 0;
            for (long multiplicity = tupleIterator.open(); multiplicity != 0; multiplicity = tupleIterator.advance()) {
                List<String> answer = new ArrayList<>();
                for (int argumentIndex = 0; argumentIndex < arity; ++argumentIndex) {
                    if (printAnswers) {
                        if (argumentIndex != 0)
                            System.out.print(" ");
                        System.out.print(
                                resourceValueFunction.apply(tupleIterator.getResource(argumentIndex)).toString());
                    }
                    answer.add(resourceValueFunction.apply(tupleIterator.getResource(argumentIndex)).toString());
                }
                if (printAnswers)
                    System.out.print("\n");
                numberOfTuples += 1;
                answers.add(answer);
            }
            if (printAnswers) {
                System.out.println("Total # of tuples: " + numberOfTuples);
                System.out.println("--------------------------------------------------------------------------------");
                System.out.println();
                System.out.println();
            }
            return answers;
        }
    }
}
